"""
In-batch loss functions for the Lightning IR framework.

This module contains loss functions that operate on batches of data,
comparing examples within the same batch for training.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import torch

from .base import InBatchLossFunction

if TYPE_CHECKING:
    from ..base import LightningIROutput
    from ..data import TrainBatch


class ScoreBasedInBatchLossFunction(InBatchLossFunction):
    """Base class for in-batch loss functions that compute in-batch indices based on scores."""

    def __init__(self, min_target_diff: float, max_num_neg_samples: int | None = None):
        """Initialize the ScoreBasedInBatchLossFunction.

        Args:
            min_target_diff (float): Minimum target difference for negative sampling.
            max_num_neg_samples (int | None): Maximum number of negative samples.
        """
        super().__init__(
            pos_sampling_technique="first",
            neg_sampling_technique="all_and_non_first",
            max_num_neg_samples=max_num_neg_samples,
        )
        self.min_target_diff = min_target_diff

    def _sort_mask(
        self, mask: torch.Tensor, num_queries: int, num_docs: int, output: LightningIROutput, batch: TrainBatch
    ) -> torch.Tensor:
        """Sort the mask based on the scores and targets.

        Args:
            mask (torch.Tensor): The initial mask tensor.
            num_queries (int): Number of queries in the batch.
            num_docs (int): Number of documents per query.
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The sorted mask tensor.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        idcs = targets.argsort(descending=True).argsort().cpu()
        idcs = idcs + torch.arange(num_queries)[:, None] * num_docs
        block_idcs = torch.arange(num_docs)[None] + torch.arange(num_queries)[:, None] * num_docs
        return mask.scatter(1, block_idcs, mask.gather(1, idcs))

    def _get_pos_mask(
        self,
        num_queries: int,
        num_docs: int,
        max_idx: torch.Tensor,
        min_idx: torch.Tensor,
        output: LightningIROutput,
        batch: TrainBatch,
    ) -> torch.Tensor:
        """Get the mask for positive samples and sort it based on scores.

        Args:
            num_queries (int): Number of queries in the batch.
            num_docs (int): Number of documents per query.
            max_idx (torch.Tensor): Maximum index for each query.
            min_idx (torch.Tensor): Minimum index for each query.
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: A mask tensor indicating the positions of positive samples, sorted by scores.
        """
        pos_mask = super()._get_pos_mask(num_queries, num_docs, max_idx, min_idx, output, batch)
        pos_mask = self._sort_mask(pos_mask, num_queries, num_docs, output, batch)
        return pos_mask

    def _get_neg_mask(
        self,
        num_queries: int,
        num_docs: int,
        max_idx: torch.Tensor,
        min_idx: torch.Tensor,
        output: LightningIROutput,
        batch: TrainBatch,
    ) -> torch.Tensor:
        """Get the mask for negative samples and sort it based on scores.

        Args:
            num_queries (int): Number of queries in the batch.
            num_docs (int): Number of documents per query.
            max_idx (torch.Tensor): Maximum index for each query.
            min_idx (torch.Tensor): Minimum index for each query.
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: A mask tensor indicating the positions of negative samples, sorted by scores.
        """
        neg_mask = super()._get_neg_mask(num_queries, num_docs, max_idx, min_idx, output, batch)
        neg_mask = self._sort_mask(neg_mask, num_queries, num_docs, output, batch)
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch).cpu()
        max_score, _ = targets.max(dim=-1, keepdim=True)
        score_diff = (max_score - targets).cpu()
        score_mask = score_diff.ge(self.min_target_diff)
        block_idcs = torch.arange(num_docs)[None] + torch.arange(num_queries)[:, None] * num_docs
        neg_mask = neg_mask.scatter(1, block_idcs, score_mask)
        # num_neg_samples might be different between queries
        num_neg_samples = neg_mask.sum(dim=1)
        min_num_neg_samples = num_neg_samples.min()
        additional_neg_samples = num_neg_samples - min_num_neg_samples
        for query_idx, neg_samples in enumerate(additional_neg_samples):
            neg_idcs = neg_mask[query_idx].nonzero().squeeze(1)
            additional_neg_idcs = neg_idcs[torch.randperm(neg_idcs.shape[0])][:neg_samples]
            assert neg_mask[query_idx, additional_neg_idcs].all()
            neg_mask[query_idx, additional_neg_idcs] = False
            assert neg_mask[query_idx].sum().eq(min_num_neg_samples)
        return neg_mask


class InBatchCrossEntropy(InBatchLossFunction):
    """In-batch cross-entropy loss function for ranking tasks.
    Originally proposed in: `Fast Single-Class Classification and the Principle of Logit Separation
    <https://arxiv.org/pdf/1705.10246v1>`_"""

    def compute_loss(self, output: LightningIROutput) -> torch.Tensor:
        """Compute the in-batch cross-entropy loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = torch.zeros(scores.shape[0], dtype=torch.long, device=scores.device)
        loss = torch.nn.functional.cross_entropy(scores, targets)
        return loss


class ScoreBasedInBatchCrossEntropy(ScoreBasedInBatchLossFunction):
    """In-batch cross-entropy loss function based on scores for ranking tasks."""

    def compute_loss(self, output: LightningIROutput) -> torch.Tensor:
        """Compute the in-batch cross-entropy loss based on scores.

        Args:
            output (LightningIROutput): The output from the model containing scores.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = torch.zeros(scores.shape[0], dtype=torch.long, device=scores.device)
        loss = torch.nn.functional.cross_entropy(scores, targets)
        return loss



"""
Listwise loss functions for the Lightning IR framework.

This module contains loss functions that operate on entire lists of items,
considering the ranking of all items simultaneously.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import torch

from .base import ListwiseLossFunction

if TYPE_CHECKING:
    from ..base import LightningIROutput
    from ..data import TrainBatch


class KLDivergence(ListwiseLossFunction):
    """Kullback-Leibler Divergence loss for listwise ranking tasks.
    Originally proposed in: `On Information and Sufficiency \
    <https://projecteuclid.org/journals/annals-of-mathematical-statistics/volume-22/issue-1/On-Information-and-Sufficiency/10.1214/aoms/1177729694.full>`_
    """

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the Kullback-Leibler Divergence loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        scores = torch.nn.functional.log_softmax(scores, dim=-1)
        targets = torch.nn.functional.log_softmax(targets.to(scores), dim=-1)
        loss = torch.nn.functional.kl_div(scores, targets, log_target=True, reduction="batchmean")
        return loss


class PearsonCorrelation(ListwiseLossFunction):
    """Pearson Correlation loss for listwise ranking tasks."""

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the Pearson Correlation loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch).to(scores)
        centered_scores = scores - scores.mean(dim=-1, keepdim=True)
        centered_targets = targets - targets.mean(dim=-1, keepdim=True)
        pearson = torch.nn.functional.cosine_similarity(centered_scores, centered_targets, dim=-1)
        loss = (1 - pearson).mean()
        return loss


class InfoNCE(ListwiseLossFunction):
    """InfoNCE loss for listwise ranking tasks.
    Originally proposed in: `Representation Learning with Contrastive Predictive Coding \
    <https://arxiv.org/abs/1807.03748>`_
    """

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the InfoNCE loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        targets = targets.argmax(dim=1)
        loss = torch.nn.functional.cross_entropy(scores, targets)
        return loss



"""Configuration and model for DPR (Dense Passage Retriever) type models. Originally proposed in \
`Dense Passage Retrieval for Open-Domain Question Answering \
<https://arxiv.org/abs/2004.04906>`_. This model type is also known as a SentenceTransformer model:
`Sentence-BERT: Sentence Embeddings using Siamese BERT-Networks \
<https://arxiv.org/abs/1908.10084>`_.
"""

from typing import Literal

import torch
from transformers import BatchEncoding

from ...bi_encoder import BiEncoderEmbedding, SingleVectorBiEncoderConfig, SingleVectorBiEncoderModel
from ...modeling_utils.embedding_post_processing import Pooler, Sparsifier


class DprConfig(SingleVectorBiEncoderConfig):
    """Configuration class for a DPR model."""

    model_type = "lir-dpr"
    """Model type for a DPR model."""

    def __init__(
        self,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        similarity_function: Literal["cosine", "dot"] = "dot",
        normalize: bool = False,
        sparsification: Literal["relu", "relu_log", "relu_2xlog"] | None = None,
        add_marker_tokens: bool = False,
        query_pooling_strategy: Literal["first", "mean", "max", "sum"] = "first",
        doc_pooling_strategy: Literal["first", "mean", "max", "sum"] = "first",
        embedding_dim: int | None = None,
        projection: Literal["linear", "linear_no_bias"] | None = "linear",
        **kwargs,
    ) -> None:
        """A DPR model encodes queries and documents separately. Before computing the similarity score, the
        contextualized token embeddings are aggregated to obtain a single embedding using a pooling strategy.
        Optionally, the pooled embeddings can be projected using a linear layer.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            similarity_function (Literal["cosine", "dot"]): Similarity function to compute scores between query and
                document embeddings. Defaults to "dot".
            normalize (bool): Whether to normalize the embeddings. Defaults to False.
            sparsification (Literal['relu', 'relu_log', 'relu_2xlog'] | None): Whether and which sparsification
                function to apply. Defaults to None.
            add_marker_tokens (bool): Whether to add marker tokens to the input sequences. Defaults to False.
            query_pooling_strategy (Literal["first", "mean", "max", "sum"]): Pooling strategy for query embeddings.
                Defaults to "first".
            doc_pooling_strategy (Literal["first", "mean", "max", "sum"]): Pooling strategy for document embeddings.
                Defaults to "first".
            embedding_dim (int | None): Dimension of the final embeddings. If None, it will be set to the hidden size
                of the backbone model. Defaults to None.
            projection (Literal["linear", "linear_no_bias"] | None): type of projection layer to apply on the pooled
                embeddings. If None, no projection is applied. Defaults to "linear".
        """
        super().__init__(
            query_length=query_length,
            doc_length=doc_length,
            similarity_function=similarity_function,
            normalize=normalize,
            sparsification=sparsification,
            add_marker_tokens=add_marker_tokens,
            query_pooling_strategy=query_pooling_strategy,
            doc_pooling_strategy=doc_pooling_strategy,
            **kwargs,
        )
        hidden_size = getattr(self, "hidden_size", None)
        if projection is None:
            self.embedding_dim = hidden_size
        else:
            self.embedding_dim = embedding_dim or hidden_size
        self.projection = projection


class DprModel(SingleVectorBiEncoderModel):
    """A single-vector DPR model. See :class:`DprConfig` for configuration options."""

    config_class = DprConfig
    """Configuration class for a DPR model."""

    def __init__(self, config: SingleVectorBiEncoderConfig, *args, **kwargs) -> None:
        """Initializes a DPR model given a :class:`DprConfig`.

        Args:
            config (SingleVectorBiEncoderConfig): Configuration for the DPR model.
        Raises:
            ValueError: If the embedding dimension is not specified in the configuration.
        """
        super().__init__(config, *args, **kwargs)
        if self.config.projection is None:
            self.projection: torch.nn.Module = torch.nn.Identity()
        else:
            if self.config.embedding_dim is None:
                raise ValueError("Unable to determine embedding dimension.")
            self.projection = torch.nn.Linear(
                self.config.hidden_size,
                self.config.embedding_dim,
                bias="no_bias" not in self.config.projection,
            )
        self.pooler = Pooler(config.pooling_strategy)
        self.sparsifier = Sparsifier(config.sparsification_strategy)

    def encode(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> BiEncoderEmbedding:
        """Encodes a batched tokenized text sequences and returns the embeddings and scoring mask.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BiEncoderEmbedding: Embeddings and scoring mask.
        """
        embeddings = self._backbone_forward(**encoding).last_hidden_state
        embeddings = self.pooler(embeddings, encoding["attention_mask"])
        embeddings = self.projection(embeddings)
        embeddings = self.sparsifier(embeddings)
        if self.config.normalize:
            embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=-1)
        return BiEncoderEmbedding(embeddings, None, encoding)



"""
Configuration module for cross-encoder models.

This module defines the configuration class used to instantiate cross-encoder models.
"""

from typing import Literal

from ..base import LightningIRConfig


class CrossEncoderConfig(LightningIRConfig):
    model_type: str = "cross-encoder"
    """Model type for cross-encoder models."""

    def __init__(
        self,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        pooling_strategy: Literal["first", "mean", "max", "sum"] = "first",
        linear_bias: bool = False,
        **kwargs,
    ):
        """Configuration class for a cross-encoder model

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            pooling_strategy (Literal['first', 'mean', 'max', 'sum']): Pooling strategy to aggregate the
                contextualized embeddings into a single vector for computing a relevance score. Defaults to "first".
            linear_bias (bool): Whether to use a bias in the prediction linear layer. Defaults to False.
        """
        super().__init__(query_length=query_length, doc_length=doc_length, **kwargs)
        self.pooling_strategy = pooling_strategy
        self.linear_bias = linear_bias



"""Torch-based Dense Searcher for Lightning IR Framework"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import torch

from ...modeling_utils.batching import _batch_pairwise_scoring
from ...models import ColConfig, DprConfig
from ..base.searcher import ExactSearchConfig, ExactSearcher
from .dense_indexer import TorchDenseIndexConfig

if TYPE_CHECKING:
    from ...bi_encoder import BiEncoderEmbedding, BiEncoderModule


class TorchDenseIndex:
    """Torch-based dense index for embeddings."""

    def __init__(self, index_dir: Path, similarity_function: Literal["dot", "cosine"], use_gpu: bool = False) -> None:
        """Initialize the TorchDenseIndex.

        Args:
            index_dir (Path): Directory where the index is stored.
            similarity_function (Literal["dot", "cosine"]): Similarity function to use for scoring.
            use_gpu (bool): Whether to use GPU for indexing. Defaults to False.
        Raises:
            ValueError: If the similarity function is not recognized.
        """
        self.index = torch.load(index_dir / "index.pt", weights_only=True)
        self.config = TorchDenseIndexConfig.from_pretrained(index_dir)
        if similarity_function == "dot":
            self.similarity_function = self.dot_similarity
        elif similarity_function == "cosine":
            self.similarity_function = self.cosine_similarity
        else:
            raise ValueError("Unknown similarity function")
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    def score(self, embeddings: torch.Tensor) -> torch.Tensor:
        """Score the embeddings against the index.

        Args:
            embeddings (torch.Tensor): The embeddings to score.
        Returns:
            torch.Tensor: The scores for the embeddings.
        """
        embeddings = embeddings.to(self.index)
        similarity = self.similarity_function(embeddings, self.index)
        return similarity

    @property
    def num_embeddings(self) -> int:
        """Get the number of embeddings in the index."""
        return self.index.shape[0]

    @staticmethod
    @_batch_pairwise_scoring
    @torch.autocast(device_type="cuda", enabled=False)
    def cosine_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute the cosine similarity between two tensors.

        Args:
            x (torch.Tensor): First tensor.
            y (torch.Tensor): Second tensor.
        Returns:
            torch.Tensor: Cosine similarity scores.
        """
        return torch.nn.functional.cosine_similarity(x[:, None], y[None], dim=-1)

    @staticmethod
    @_batch_pairwise_scoring
    @torch.autocast(device_type="cuda", enabled=False)
    def dot_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute the dot product similarity between two tensors.

        Args:
            x (torch.Tensor): First tensor.
            y (torch.Tensor): Second tensor.
        Returns:
            torch.Tensor: Dot product similarity scores.
        """
        return torch.matmul(x, y.T)

    def to_gpu(self) -> None:
        """Convert the index to GPU format."""
        self.index = self.index.to(self.device)


class TorchDenseSearcher(ExactSearcher):
    """Torch-based dense searcher for embeddings."""

    def __init__(
        self,
        index_dir: Path,
        search_config: TorchDenseSearchConfig,
        module: BiEncoderModule,
        use_gpu: bool = True,
    ) -> None:
        """Initialize the TorchDenseSearcher.

        Args:
            index_dir (Path): Directory where the index is stored.
            search_config (TorchDenseSearchConfig): Configuration for the dense search.
            module (BiEncoderModule): Bi-encoder module to use for searching.
            use_gpu (bool): Whether to use GPU for searching. Defaults to True.
        """
        self.search_config: TorchDenseSearchConfig
        self.index = TorchDenseIndex(index_dir, module.config.similarity_function, use_gpu)
        super().__init__(index_dir, search_config, module, use_gpu)
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    def to_gpu(self) -> None:
        """Move the searcher to the GPU if available."""
        super().to_gpu()
        self.index.to_gpu()

    def _score(self, query_embeddings: BiEncoderEmbedding) -> torch.Tensor:
        if query_embeddings.scoring_mask is None:
            embeddings = query_embeddings.embeddings[:, 0]
        else:
            embeddings = query_embeddings.embeddings[query_embeddings.scoring_mask]
        scores = self.index.score(embeddings)
        return scores


class TorchDenseSearchConfig(ExactSearchConfig):
    """Configuration for the TorchDenseSearcher."""

    search_class = TorchDenseSearcher
    SUPPORTED_MODELS = {ColConfig.model_type, DprConfig.model_type}



"""
Module for indexing and searching using Seismic.

This module provides the classes and configurations for indexing and searching using Seismic in the Lightning IR
framework.

Seismic: `Efficient Inverted Indexes for Approximate Retrieval over Learned Sparse Representations
<https://dl.acm.org/doi/10.1145/3626772.3657769>`_
"""

from .seismic_indexer import SeismicIndexConfig, SeismicIndexer
from .seismic_searcher import SeismicSearchConfig, SeismicSearcher

__all__ = [
    "SeismicIndexConfig",
    "SeismicIndexer",
    "SeismicSearcher",
    "SeismicSearchConfig",
]



import codecs
import json
from pathlib import Path
from typing import Any, Literal, NamedTuple

import ir_datasets
from ir_datasets.datasets.base import Dataset
from ir_datasets.datasets.nano_beir import parquet_iter
from ir_datasets.formats import (
    BaseDocPairs,
    BaseDocs,
    BaseQrels,
    BaseQueries,
    BaseScoredDocs,
    GenericScoredDoc,
    jsonl,
    trec,
    tsv,
)
from ir_datasets.util import Cache, DownloadConfig

CONSTITUENT_TYPE_MAP: dict[str, dict[str, type]] = {
    "docs": {
        ".json": jsonl.JsonlDocs,
        ".jsonl": jsonl.JsonlDocs,
        ".tsv": tsv.TsvDocs,
    },
    "queries": {
        ".json": jsonl.JsonlQueries,
        ".jsonl": jsonl.JsonlQueries,
        ".tsv": tsv.TsvQueries,
    },
    "qrels": {".tsv": trec.TrecQrels, ".qrels": trec.TrecQrels},
    "scoreddocs": {".run": trec.TrecScoredDocs, ".tsv": trec.TrecScoredDocs},
    "docpairs": {".tsv": tsv.TsvDocPairs},
}


def _load_constituent(
    dataset_id: str,
    constituent: Path | str | dict[str, Any] | None,
    constituent_type: Literal["docs", "queries", "qrels", "scoreddocs", "docpairs"] | type,
    **kwargs,
) -> Any:
    if constituent is None:
        return None
    if isinstance(constituent, dict):
        constituent_path = Path(constituent["cache_path"])
        cache = _register_and_get_cache(dataset_id, constituent)
    elif constituent in ir_datasets.registry._registered:
        return getattr(ir_datasets.load(constituent), f"{constituent_type}_handler")()
    else:
        constituent_path = Path(constituent)
        cache = Cache(None, constituent_path)
        if not constituent_path.exists():
            raise ValueError(f"unable to load {constituent}, expected an `ir_datasets` id or valid path")
    if isinstance(constituent_type, str):
        suffix = constituent_path.suffixes[0]
        constituent_types = CONSTITUENT_TYPE_MAP[constituent_type]
        if suffix not in constituent_types:
            raise ValueError(f"Unknown file type: {suffix}, expected one of {constituent_types.keys()}")
        ConstituentType = constituent_types[suffix]
    else:
        ConstituentType = constituent_type
    return ConstituentType(cache, **kwargs)


def _register_and_get_cache(dataset_id: str, dlc_contents: dict[str, Any]) -> Cache:
    extractors = dlc_contents.pop("extractors", [])
    base_id = dataset_id.split("/")[0]
    new_id = dataset_id.removeprefix(base_id + "/")
    base_path = ir_datasets.util.home_path()
    dlc = DownloadConfig.context(base_id, base_path / base_id)
    dlc.contents()[new_id] = dlc_contents
    dataset_dlc = dlc[new_id]
    file_path = Path(dlc_contents["cache_path"])
    for extractor in extractors:
        dataset_dlc = extractor(dataset_dlc)
    return Cache(dataset_dlc, base_path / file_path)


def register_new_dataset(
    dataset_id: str,
    docs: Path | str | dict[str, str] | None = None,
    DocsType: type[BaseDocs] | None = None,
    queries: Path | str | dict[str, str] | None = None,
    QueriesType: type[BaseQueries] | None = None,
    qrels: Path | str | dict[str, str] | None = None,
    QrelsType: type[BaseQrels] | None = None,
    docpairs: Path | str | dict[str, str] | None = None,
    DocpairsType: type[BaseDocPairs] | None = None,
    scoreddocs: Path | str | dict[str, str] | None = None,
    ScoreddocsType: type[BaseScoredDocs] | None = None,
    qrels_defs: dict[int, str] | None = None,
):
    if dataset_id in ir_datasets.registry._registered:
        return

    docs = _load_constituent(dataset_id, docs, DocsType or "docs")
    queries = _load_constituent(dataset_id, queries, QueriesType or "queries")
    qrels = _load_constituent(
        dataset_id, qrels, QrelsType or "qrels", qrels_defs=qrels_defs if qrels_defs is not None else {}
    )
    docpairs = _load_constituent(dataset_id, docpairs, DocpairsType or "docpairs")
    scoreddocs = _load_constituent(dataset_id, scoreddocs, ScoreddocsType or "scoreddocs")

    ir_datasets.registry.register(dataset_id, Dataset(docs, queries, qrels, docpairs, scoreddocs))


class ScoredDocTuple(NamedTuple):
    query_id: str
    doc_ids: tuple[str, ...]
    scores: tuple[float, ...] | None
    num_docs: int


class ScoredDocTuples(BaseDocPairs):
    def __init__(self, docpairs_dlc):
        self._docpairs_dlc = docpairs_dlc

    def docpairs_path(self):
        return self._docpairs_dlc.path()

    def docpairs_iter(self):
        file_type = None
        if self._docpairs_dlc.path().suffix == ".json":
            file_type = "json"
        elif self._docpairs_dlc.path().suffix in (".tsv", ".run"):
            file_type = "tsv"
        else:
            raise ValueError(f"Unknown file type: {self._docpairs_dlc.path().suffix}")
        with self._docpairs_dlc.stream() as f:
            f = codecs.getreader("utf8")(f)
            for line in f:
                if file_type == "json":
                    data = json.loads(line)
                    qid, *doc_data = data
                    pids, scores = zip(*doc_data)
                    pids = tuple(str(pid) for pid in pids)
                else:
                    cols = line.rstrip().split()
                    pos_score, neg_score, qid, pid1, pid2 = cols
                    pids = (pid1, pid2)
                    scores = (float(pos_score), float(neg_score))
                yield ScoredDocTuple(str(qid), pids, scores, len(pids))

    def docpairs_cls(self):
        return ScoredDocTuple


class ParquetScoredDocs(BaseScoredDocs):
    def __init__(self, scoreddocs_dlc, negate_score=False):
        self._scoreddocs_dlc = scoreddocs_dlc

    def scoreddocs_path(self):
        return self._scoreddocs_dlc.path()

    def scoreddocs_iter(self):
        for d in parquet_iter(self._scoreddocs_dlc.path()):
            yield GenericScoredDoc(str(d["qid"]), str(d["docno"]), float(d["score"]))



"""
Approximate ranking loss functions for the Lightning IR framework.

This module contains loss functions that use approximation techniques to compute
ranking-based metrics like NDCG, MRR, and RankMSE in a differentiable manner.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import torch

from .base import ListwiseLossFunction

if TYPE_CHECKING:
    from ..base import LightningIROutput
    from ..data import TrainBatch


class ApproxLossFunction(ListwiseLossFunction):
    """Base class for approximate loss functions that compute ranks from scores."""

    def __init__(self, temperature: float = 1) -> None:
        """Initialize the ApproxLossFunction.

        Args:
            temperature (float): Temperature parameter for scaling the scores. Defaults to 1.
        """
        super().__init__()
        self.temperature = temperature

    @staticmethod
    def get_approx_ranks(scores: torch.Tensor, temperature: float) -> torch.Tensor:
        """Compute approximate ranks from scores.

        Args:
            scores (torch.Tensor): The input scores.
            temperature (float): Temperature parameter for scaling the scores.
        Returns:
            torch.Tensor: The computed approximate ranks.
        """
        score_diff = scores[:, None] - scores[..., None]
        normalized_score_diff = torch.sigmoid(score_diff / temperature)
        # set diagonal to 0
        normalized_score_diff = normalized_score_diff * (1 - torch.eye(scores.shape[1], device=scores.device))
        approx_ranks = normalized_score_diff.sum(-1) + 1
        return approx_ranks


class ApproxNDCG(ApproxLossFunction):
    """Approximate NDCG loss function for ranking tasks.
    Originally proposed in: `Cumulated Gain-Based Evaluation of IR Techniques \
    <https://dl.acm.org/doi/10.1145/582415.582418>`_"""

    def __init__(self, temperature: float = 1, scale_gains: bool = True):
        """Initialize the ApproxNDCG loss function.

        Args:
            temperature (float): Temperature parameter for scaling the scores. Defaults to 1.
            scale_gains (bool): Whether to scale the gains. Defaults to True.
        """
        super().__init__(temperature)
        self.scale_gains = scale_gains

    @staticmethod
    def get_dcg(
        ranks: torch.Tensor,
        targets: torch.Tensor,
        k: int | None = None,
        scale_gains: bool = True,
    ) -> torch.Tensor:
        """Compute the Discounted Cumulative Gain (DCG) for the given ranks and targets.

        Args:
            ranks (torch.Tensor): The ranks of the items.
            targets (torch.Tensor): The relevance scores of the items.
            k (int | None): Optional cutoff for the ranks. If provided, only computes DCG for the top k items.
            scale_gains (bool): Whether to scale the gains. Defaults to True.
        Returns:
            torch.Tensor: The computed DCG values.
        """
        log_ranks = torch.log2(1 + ranks)
        discounts = 1 / log_ranks
        if scale_gains:
            gains = 2**targets - 1
        else:
            gains = targets
        dcgs = gains * discounts
        if k is not None:
            dcgs = dcgs.masked_fill(ranks > k, 0)
        return dcgs.sum(dim=-1)

    @staticmethod
    def get_ndcg(
        ranks: torch.Tensor,
        targets: torch.Tensor,
        k: int | None = None,
        scale_gains: bool = True,
        optimal_targets: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Compute the Normalized Discounted Cumulative Gain (NDCG) for the given ranks and targets.

        Args:
            ranks (torch.Tensor): The ranks of the items.
            targets (torch.Tensor): The relevance scores of the items.
            k (int | None): Cutoff for the ranks. If provided, only computes NDCG for the top k items. Defaults to None.
            scale_gains (bool): Whether to scale the gains. Defaults to True.
            optimal_targets (torch.Tensor | None): Optional tensor of optimal targets for normalization. If None, uses
                the targets. Defaults to None.
        Returns:
            torch.Tensor: The computed NDCG values.
        """
        targets = targets.clamp(min=0)
        if optimal_targets is None:
            optimal_targets = targets
        optimal_ranks = torch.argsort(torch.argsort(optimal_targets, descending=True))
        optimal_ranks = optimal_ranks + 1
        dcg = ApproxNDCG.get_dcg(ranks, targets, k, scale_gains)
        idcg = ApproxNDCG.get_dcg(optimal_ranks, optimal_targets, k, scale_gains)
        ndcg = dcg / (idcg.clamp(min=1e-12))
        return ndcg

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the ApproxNDCG loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        ndcg = self.get_ndcg(approx_ranks, targets, k=None, scale_gains=self.scale_gains)
        loss = 1 - ndcg
        return loss.mean()


class ApproxMRR(ApproxLossFunction):
    """Approximate Mean Reciprocal Rank (MRR) loss function for ranking tasks."""

    def __init__(self, temperature: float = 1):
        """Initialize the ApproxMRR loss function.

        Args:
            temperature (float): Temperature parameter for scaling the scores. Defaults to 1.
        """
        super().__init__(temperature)

    @staticmethod
    def get_mrr(ranks: torch.Tensor, targets: torch.Tensor, k: int | None = None) -> torch.Tensor:
        """Compute the Mean Reciprocal Rank (MRR) for the given ranks and targets.

        Args:
            ranks (torch.Tensor): The ranks of the items.
            targets (torch.Tensor): The relevance scores of the items.
            k (int | None): Optional cutoff for the ranks. If provided, only computes MRR for the top k items.
        Returns:
            torch.Tensor: The computed MRR values.
        """
        targets = targets.clamp(None, 1)
        reciprocal_ranks = 1 / ranks
        mrr = reciprocal_ranks * targets
        if k is not None:
            mrr = mrr.masked_fill(ranks > k, 0)
        mrr = mrr.max(dim=-1)[0]
        return mrr

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the ApproxMRR loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        mrr = self.get_mrr(approx_ranks, targets, k=None)
        loss = 1 - mrr
        return loss.mean()


class ApproxRankMSE(ApproxLossFunction):
    """Approximate Rank Mean Squared Error (RankMSE) loss function for ranking tasks.
    Originally proposed in: `Rank-DistiLLM: Closing the Effectiveness Gap Between Cross-Encoders and LLMs
    for Passage Re-ranking <https://link.springer.com/chapter/10.1007/978-3-031-88714-7_31>`_"""

    def __init__(
        self,
        temperature: float = 1,
        discount: Literal["log2", "reciprocal"] | None = None,
    ):
        """Initialize the ApproxRankMSE loss function.

        Args:
            temperature (float): Temperature parameter for scaling the scores. Defaults to 1.
            discount (Literal["log2", "reciprocal"] | None): Discounting strategy for the loss. If None, no discounting
                is applied. Defaults to None.
        """
        super().__init__(temperature)
        self.discount = discount

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the ApproxRankMSE loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        ranks = torch.argsort(torch.argsort(targets, descending=True)) + 1
        loss = torch.nn.functional.mse_loss(approx_ranks, ranks.to(approx_ranks), reduction="none")
        if self.discount == "log2":
            weight = 1 / torch.log2(ranks + 1)
        elif self.discount == "reciprocal":
            weight = 1 / ranks
        else:
            weight = 1
        loss = loss * weight
        loss = loss.mean()
        return loss



"""Base searcher class and configuration for retrieval tasks."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import torch

from ...bi_encoder.bi_encoder_model import BiEncoderEmbedding, SingleVectorBiEncoderConfig
from .packed_tensor import PackedTensor

if TYPE_CHECKING:
    from ...bi_encoder import BiEncoderModule, BiEncoderOutput


def cat_arange(arange_starts: torch.Tensor, arange_ends: torch.Tensor) -> torch.Tensor:
    """Concatenates arange tensors into a single tensor.

    Args:
        arange_starts (torch.Tensor): The start indices of the ranges.
        arange_ends (torch.Tensor): The end indices of the ranges.
    Returns:
        torch.Tensor: A tensor containing the concatenated ranges.
    """
    arange_lengths = arange_ends - arange_starts
    offsets = torch.cumsum(arange_lengths, dim=0) - arange_lengths - arange_starts
    return torch.arange(arange_lengths.sum()) - torch.repeat_interleave(offsets, arange_lengths)


class Searcher(ABC):
    """Base class for searchers in the Lightning IR framework."""

    def __init__(
        self, index_dir: Path | str, search_config: SearchConfig, module: BiEncoderModule, use_gpu: bool = True
    ) -> None:
        """Initialize the Searcher.

        Args:
            index_dir (Path | str): The directory containing the index files.
            search_config (SearchConfig): The configuration for the search.
            module (BiEncoderModule): The bi-encoder module to use for scoring.
            use_gpu (bool): Whether to use GPU for computations. Defaults to True.
        Raises:
            ValueError: If the document lengths do not match the index.
        """
        super().__init__()
        self.index_dir = Path(index_dir)
        self.search_config = search_config
        self.use_gpu = use_gpu
        self.module = module
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

        self.doc_ids = (self.index_dir / "doc_ids.txt").read_text().split()
        self.doc_lengths = torch.load(self.index_dir / "doc_lengths.pt", weights_only=True)

        self.to_gpu()

        self.num_docs = len(self.doc_ids)
        self.cumulative_doc_lengths = torch.cumsum(self.doc_lengths, dim=0)
        self.num_embeddings = int(self.cumulative_doc_lengths[-1].item())

        self.doc_is_single_vector = self.num_docs == self.num_embeddings
        self.query_is_single_vector = isinstance(module.config, SingleVectorBiEncoderConfig) or getattr(
            module.config, "query_pooling_strategy", None
        ) in {"first", "mean", "min", "max"}

        if self.doc_lengths.shape[0] != self.num_docs or self.doc_lengths.sum() != self.num_embeddings:
            raise ValueError("doc_lengths do not match index")

    def to_gpu(self) -> None:
        """Move the searcher to the GPU if available."""
        self.doc_lengths = self.doc_lengths.to(self.device)

    def _filter_and_sort(
        self, doc_scores: PackedTensor, doc_idcs: PackedTensor, k: int | None = None
    ) -> tuple[PackedTensor, PackedTensor]:
        """Filter and sort the document scores and indices.

        Args:
            doc_scores (PackedTensor): The document scores.
            doc_idcs (PackedTensor): The document indices.
            k (int | None): The number of top documents to return. If None, use the configured k from search_config.
                Defaults to None.
        Returns:
            tuple[PackedTensor, PackedTensor]: The filtered and sorted document scores and indices.
        """
        k = k or self.search_config.k
        per_query_doc_scores = torch.split(doc_scores, doc_scores.lengths)
        per_query_doc_idcs = torch.split(doc_idcs, doc_idcs.lengths)
        num_docs = []
        new_doc_scores = []
        new_doc_idcs = []
        for _scores, _idcs in zip(per_query_doc_scores, per_query_doc_idcs):
            _k = min(k, _scores.shape[0])
            top_values, top_idcs = torch.topk(_scores, _k)
            new_doc_scores.append(top_values)
            new_doc_idcs.append(_idcs[top_idcs])
            num_docs.append(_k)
        return PackedTensor(torch.cat(new_doc_scores), lengths=num_docs), PackedTensor(
            torch.cat(new_doc_idcs), lengths=num_docs
        )

    @abstractmethod
    def search(self, output: BiEncoderOutput) -> tuple[PackedTensor, list[list[str]]]:
        """Search for documents based on the output of the bi-encoder model.

        Args:
            output (BiEncoderOutput): The output from the bi-encoder model containing query and document embeddings.
        Returns:
            tuple[PackedTensor, list[list[str]]]: The top-k scores and corresponding document IDs.
        """
        ...


class ExactSearcher(Searcher):
    """Searcher that retrieves documents using exact matching of query embeddings."""

    def search(self, output: BiEncoderOutput) -> tuple[PackedTensor, list[list[str]]]:
        """Search for documents based on the output of the bi-encoder model.

        Args:
            output (BiEncoderOutput): The output from the bi-encoder model containing query and document embeddings.
        Returns:
            tuple[PackedTensor, list[list[str]]]: The top-k scores and corresponding document IDs.
        """
        query_embeddings = output.query_embeddings
        if query_embeddings is None:
            raise ValueError("Expected query_embeddings in BiEncoderOutput")
        query_embeddings = query_embeddings.to(self.device)

        scores = self._score(query_embeddings)

        # aggregate doc token scores
        if not self.doc_is_single_vector:
            scores = torch.scatter_reduce(
                torch.zeros(scores.shape[0], self.num_docs, device=scores.device),
                1,
                self.doc_token_idcs[None].long().expand_as(scores),
                scores,
                "amax",
            )

        # aggregate query token scores
        if not self.query_is_single_vector:
            if query_embeddings.scoring_mask is None:
                raise ValueError("Expected scoring_mask in multi-vector query_embeddings")
            query_lengths = query_embeddings.scoring_mask.sum(-1)
            query_token_idcs = torch.arange(query_lengths.shape[0]).to(query_lengths).repeat_interleave(query_lengths)
            scores = torch.scatter_reduce(
                torch.zeros(query_lengths.shape[0], self.num_docs, device=scores.device),
                0,
                query_token_idcs[:, None].expand_as(scores),
                scores,
                self.module.config.query_aggregation_function,
            )
        top_scores, top_idcs = torch.topk(scores, self.search_config.k)
        doc_ids = [[self.doc_ids[idx] for idx in _doc_idcs] for _doc_idcs in top_idcs.tolist()]
        return PackedTensor(top_scores.view(-1), lengths=[self.search_config.k] * len(doc_ids)), doc_ids

    @property
    def doc_token_idcs(self) -> torch.Tensor:
        """Get the document token indices for scoring.

        Returns:
            torch.Tensor: The document token indices.
        """
        if not hasattr(self, "_doc_token_idcs"):
            self._doc_token_idcs = (
                torch.arange(self.doc_lengths.shape[0])
                .to(device=self.doc_lengths.device)
                .repeat_interleave(self.doc_lengths)
            )
        return self._doc_token_idcs

    @abstractmethod
    def _score(self, query_embeddings: BiEncoderEmbedding) -> torch.Tensor: ...


class ApproximateSearcher(Searcher):
    def search(self, output: BiEncoderOutput) -> tuple[PackedTensor, list[list[str]]]:
        """Search for documents based on the output of the bi-encoder model.

        Args:
            output (BiEncoderOutput): The output from the bi-encoder model containing query and document embeddings.
        Returns:
            tuple[PackedTensor, list[list[str]]]: The top-k scores and corresponding document IDs.
        """
        query_embeddings = output.query_embeddings
        if query_embeddings is None:
            raise ValueError("Expected query_embeddings in BiEncoderOutput")
        query_embeddings = query_embeddings.to(self.device)

        candidate_scores, candidate_idcs = self._candidate_retrieval(query_embeddings)
        scores, doc_idcs = self._aggregate_doc_scores(candidate_scores, candidate_idcs, query_embeddings)
        scores = self._aggregate_query_scores(scores, query_embeddings)
        scores, doc_idcs = self._filter_and_sort(scores, doc_idcs)
        doc_ids = [
            [self.doc_ids[doc_idx] for doc_idx in _doc_ids.tolist()] for _doc_ids in doc_idcs.split(doc_idcs.lengths)
        ]

        return scores, doc_ids

    def _aggregate_doc_scores(
        self, candidate_scores: PackedTensor, candidate_idcs: PackedTensor, query_embeddings: BiEncoderEmbedding
    ) -> tuple[PackedTensor, PackedTensor]:
        if self.doc_is_single_vector:
            return candidate_scores, candidate_idcs

        query_lengths = query_embeddings.scoring_mask.sum(-1)
        num_query_vecs = query_lengths.sum()

        # map vec_idcs to doc_idcs
        candidate_doc_idcs = torch.searchsorted(
            self.cumulative_doc_lengths,
            candidate_idcs.to(self.cumulative_doc_lengths.device),
            side="right",
        )

        # convert candidate_scores `num_query_vecs x candidate_k` to `num_query_doc_pairs x num_query_vecs`
        # and aggregate the maximum doc_vector score per query_vector
        max_query_length = query_lengths.max()
        num_docs_per_query_candidate = torch.tensor(candidate_scores.lengths)

        query_idcs = (
            torch.arange(query_lengths.shape[0], device=query_lengths.device)
            .repeat_interleave(query_lengths)
            .repeat_interleave(num_docs_per_query_candidate)
        )
        query_vector_idcs = cat_arange(torch.zeros_like(query_lengths), query_lengths).repeat_interleave(
            num_docs_per_query_candidate
        )

        stacked = torch.stack([query_idcs, candidate_doc_idcs])
        unique_idcs, ranking_doc_idcs = stacked.unique(return_inverse=True, dim=1)
        num_docs = unique_idcs[0].bincount()
        doc_idcs = PackedTensor(unique_idcs[1], lengths=num_docs.tolist())
        total_num_docs = num_docs.sum()

        unpacked_scores = torch.full((total_num_docs * max_query_length,), float("nan"), device=query_lengths.device)
        index = ranking_doc_idcs * max_query_length + query_vector_idcs
        unpacked_scores = torch.scatter_reduce(
            unpacked_scores, 0, index, candidate_scores, "max", include_self=False
        ).view(total_num_docs, max_query_length)

        # impute the missing values
        if self.search_config.imputation_strategy == "gather":
            # reconstruct the doc embeddings and re-compute the scores
            imputation_values = torch.empty_like(unpacked_scores)
            doc_embeddings = self._reconstruct_doc_embeddings(doc_idcs)
            similarity = self.module.model.compute_similarity(query_embeddings, doc_embeddings, doc_idcs.lengths)
            unpacked_scores = self.module.model._aggregate(
                similarity, doc_embeddings.scoring_mask, "max", dim=-1
            ).squeeze(-1)
        elif self.search_config.imputation_strategy == "min":
            per_query_vec_min = torch.scatter_reduce(
                torch.empty(num_query_vecs),
                0,
                torch.arange(query_lengths.sum()).repeat_interleave(num_docs_per_query_candidate),
                candidate_scores,
                "min",
                include_self=False,
            )
            imputation_values = torch.nn.utils.rnn.pad_sequence(
                per_query_vec_min.split(query_lengths.tolist()), batch_first=True
            ).repeat_interleave(num_docs, dim=0)
        elif self.search_config.imputation_strategy == "zero":
            imputation_values = torch.zeros_like(unpacked_scores)
        else:
            raise ValueError(f"Invalid imputation strategy: {self.search_config.imputation_strategy}")

        is_nan = torch.isnan(unpacked_scores)
        unpacked_scores[is_nan] = imputation_values[is_nan]

        return PackedTensor(unpacked_scores, lengths=num_docs.tolist()), doc_idcs

    def _aggregate_query_scores(self, scores: PackedTensor, query_embeddings: BiEncoderEmbedding) -> PackedTensor:
        if self.query_is_single_vector:
            return scores
        query_scoring_mask = query_embeddings.scoring_mask.repeat_interleave(torch.tensor(scores.lengths), dim=0)
        scores = PackedTensor(
            self.module.model._aggregate(
                scores, query_scoring_mask, self.module.config.query_aggregation_function, dim=1
            ).squeeze(-1),
            lengths=scores.lengths,
        )
        return scores

    @abstractmethod
    def _candidate_retrieval(self, query_embeddings: BiEncoderEmbedding) -> tuple[PackedTensor, PackedTensor]:
        """Retrieves initial candidates using the query embeddings. Returns candidate scores and candidate vector
        indices of shape `num_query_vecs x candidate_k` (packed). Candidate indices are None if all doc vectors are
        scored.

        Args:
            query_embeddings (BiEncoderEmbedding): The query embeddings to use for candidate retrieval.
        Returns:
            tuple[PackedTensor, PackedTensor]: The candidate scores and candidate vector indices.
        """
        ...

    @abstractmethod
    def _gather_doc_embeddings(self, idcs: torch.Tensor) -> torch.Tensor:
        """Gather document embeddings based on the provided indices.

        Args:
            idcs (torch.Tensor): The indices of the document embeddings to gather.
        Returns:
            torch.Tensor: The gathered document embeddings.
        """
        ...

    def _reconstruct_doc_embeddings(self, doc_idcs: PackedTensor) -> BiEncoderEmbedding:
        """Reconstruct document embeddings based on the provided document indices.

        Args:
            doc_idcs (PackedTensor): The packed tensor containing document indices.
        Returns:
            BiEncoderEmbedding: The reconstructed document embeddings.
        """
        # unique doc_idcs per query
        unique_doc_idcs, inverse_idcs = torch.unique(doc_idcs, return_inverse=True)

        # gather all vectors for unique doc_idcs
        doc_lengths = self.doc_lengths[unique_doc_idcs]
        start_doc_idcs = self.cumulative_doc_lengths[unique_doc_idcs - 1]
        start_doc_idcs[unique_doc_idcs == 0] = 0
        all_doc_idcs = cat_arange(start_doc_idcs, start_doc_idcs + doc_lengths)
        all_doc_embeddings = self._gather_doc_embeddings(all_doc_idcs)
        unique_embeddings = torch.nn.utils.rnn.pad_sequence(
            list(torch.split(all_doc_embeddings, doc_lengths.tolist())),
            batch_first=True,
        ).to(inverse_idcs.device)
        embeddings = unique_embeddings[inverse_idcs]

        # mask out padding
        doc_lengths = doc_lengths[inverse_idcs]
        scoring_mask = torch.arange(embeddings.shape[1], device=embeddings.device) < doc_lengths[:, None]
        doc_embeddings = BiEncoderEmbedding(embeddings=embeddings, scoring_mask=scoring_mask, encoding=None)
        return doc_embeddings


class SearchConfig:
    """Configuration class for searchers in the Lightning IR framework."""

    search_class: type[Searcher]

    SUPPORTED_MODELS: set[str]

    def __init__(self, k: int = 10) -> None:
        """Initialize the SearchConfig.

        Args:
            k (int): The number of top documents to retrieve. Defaults to 10.
        """
        self.k = k


class ExactSearchConfig(SearchConfig):
    """Configuration class for exact searchers in the Lightning IR framework."""

    search_class = ExactSearcher


class ApproximateSearchConfig(SearchConfig):
    """Configuration class for approximate searchers in the Lightning IR framework."""

    search_class = ApproximateSearcher

    def __init__(
        self, k: int = 10, candidate_k: int = 100, imputation_strategy: Literal["min", "gather", "zero"] = "gather"
    ) -> None:
        """Initialize the ApproximateSearchConfig.

        Args:
            k (int): The number of top documents to retrieve. Defaults to 10.
            candidate_k (int): The number of candidate documents to consider for scoring. Defaults to 100.
            imputation_strategy (Literal["min", "gather", "zero"]): Strategy for imputing missing scores. Defaults to
                "gather".
        """
        super().__init__(k)
        self.k = k
        self.candidate_k = candidate_k
        self.imputation_strategy = imputation_strategy



"""Configuration, model, and tokenizer for MVR (Multi-View Representation) type models. Originally proposed in
`Multi-View Document Representation Learning for Open-Domain Dense Retrieval \
<https://aclanthology.org/2022.acl-long.414/>`_.
"""

from typing import Literal

import torch
from tokenizers.processors import TemplateProcessing
from transformers import BatchEncoding

from lightning_ir.bi_encoder.bi_encoder_model import BiEncoderEmbedding

from ...bi_encoder import BiEncoderTokenizer, MultiVectorBiEncoderConfig, MultiVectorBiEncoderModel
from ...modeling_utils.embedding_post_processing import Pooler


class MvrConfig(MultiVectorBiEncoderConfig):
    """Configuration class for a MVR model."""

    model_type = "mvr"
    """Model type for a MVR model."""

    def __init__(
        self,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        similarity_function: Literal["cosine", "dot"] = "dot",
        normalization_strategy: Literal["l2"] | None = None,
        add_marker_tokens: bool = False,
        embedding_dim: int | None = None,
        projection: Literal["linear", "linear_no_bias"] | None = "linear",
        num_viewer_tokens: int | None = 8,
        **kwargs,
    ):
        """A MVR model encodes queries and document separately. It uses a single vector to represent the query and
        multiple vectors to represent the document. The document representation is obtained from n viewer tokens ([VIE])
        prepended to the document. During training, a contrastive loss pushes the viewer token representations away
        from one another, such that they represent different "views" of the document. Only the maximum similarity
        between the query vector and the viewer token vectors is used to compute the relevance score.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            similarity_function (Literal['cosine', 'dot']): Similarity function to compute scores between query and
                document embeddings. Defaults to "dot".
            normalization_strategy (Literal['l2'] | None): Whether to normalize query and document embeddings.
                Defaults to None.
            add_marker_tokens (bool): Whether to prepend extra marker tokens [Q] / [D] to queries / documents.
                Defaults to False.
            embedding_dim (int | None): Dimension of the final embeddings. If None, it will be set to the hidden size
                of the backbone model. Defaults to None.
            projection (Literal["linear", "linear_no_bias"] | None): type of projection layer to apply on the pooled
                embeddings. If None, no projection is applied. Defaults to "linear".
            num_viewer_tokens (int | None): Number of viewer tokens to prepend to the document. Defaults to 8.
        """
        super().__init__(
            query_length=query_length,
            doc_length=doc_length,
            similarity_function=similarity_function,
            normalization_strategy=normalization_strategy,
            add_marker_tokens=add_marker_tokens,
            embedding_dim=embedding_dim,
            projection=projection,
            **kwargs,
        )
        self.num_viewer_tokens = num_viewer_tokens


class MvrModel(MultiVectorBiEncoderModel):
    """MVR model for multi-view representation learning."""

    config_class = MvrConfig
    """Configuration class for MVR models."""

    def __init__(self, config: MvrConfig, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        """Initialize a MVR model.

        Args:
            config (MvrConfig): Configuration for the MVR model.
        """
        if self.config.projection is None:
            self.projection: torch.nn.Module = torch.nn.Identity()
        else:
            if self.config.embedding_dim is None:
                raise ValueError("Unable to determine embedding dimension.")
            self.projection = torch.nn.Linear(
                self.config.hidden_size,
                self.config.embedding_dim,
                bias="no_bias" not in self.config.projection,
            )
        self.query_pooler = Pooler("first")

    def scoring_mask(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> torch.Tensor:
        """Computes a scoring mask for batched tokenized text sequences which is used in the scoring function to mask
        out vectors during scoring.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            torch.Tensor: Scoring mask.
        """
        if input_type == "query":
            return torch.ones(encoding.input_ids.shape[0], 1, dtype=torch.bool, device=encoding.input_ids.device)
        elif input_type == "doc":
            return torch.ones(
                encoding.input_ids.shape[0],
                self.config.num_viewer_tokens,
                dtype=torch.bool,
                device=encoding.input_ids.device,
            )
        else:
            raise ValueError(f"Invalid input type: {input_type}")

    def encode(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> BiEncoderEmbedding:
        """Encodes a batched tokenized text sequences and returns the embeddings and scoring mask.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BiEncoderEmbedding: Embeddings and scoring mask.
        """
        embeddings = self._backbone_forward(**encoding).last_hidden_state
        embeddings = self.projection(embeddings)
        if input_type == "query":
            embeddings = self.query_pooler(embeddings, None)
        elif input_type == "doc":
            embeddings = embeddings[:, 1 : self.config.num_viewer_tokens + 1]
        else:
            raise ValueError(f"Invalid input type: {input_type}")
        if self.config.normalization_strategy == "l2":
            embeddings = torch.nn.functional.normalize(embeddings, dim=-1)
        scoring_mask = self.scoring_mask(encoding, input_type)
        return BiEncoderEmbedding(embeddings, scoring_mask, encoding)


class MvrTokenizer(BiEncoderTokenizer):
    config_class = MvrConfig

    def __init__(
        self,
        *args,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        add_marker_tokens: bool = False,
        num_viewer_tokens: int = 8,
        **kwargs,
    ):
        super().__init__(
            *args,
            query_length=query_length,
            doc_length=doc_length,
            add_marker_tokens=add_marker_tokens,
            num_viewer_tokens=num_viewer_tokens,
            **kwargs,
        )
        self.num_viewer_tokens = num_viewer_tokens
        if num_viewer_tokens is not None:
            viewer_tokens = [f"[VIE{idx}]" for idx in range(num_viewer_tokens)]
            self.add_tokens(viewer_tokens, special_tokens=True)
            special_tokens = [
                ("[CLS]", self.cls_token_id),
                ("[SEP]", self.sep_token_id),
            ] + [
                (viewer_tokens[viewer_token_id], self.viewer_token_id(viewer_token_id))
                for viewer_token_id in range(num_viewer_tokens)
            ]
            viewer_tokens_string = " ".join(viewer_tokens)
            if self.doc_token_id is not None:
                prefix = f"[CLS] {self.DOC_TOKEN}"
                special_tokens.append((self.DOC_TOKEN, self.doc_token_id))
            else:
                prefix = "[CLS]"
            self.doc_post_processor = TemplateProcessing(
                single=f"{prefix} {viewer_tokens_string} $0 [SEP]",
                pair="[CLS] $A [SEP] $B:1 [SEP]:1",
                special_tokens=special_tokens,
            )

    def viewer_token_id(self, viewer_token_id: int) -> int | None:
        """The token id of the query token if marker tokens are added.

        :return: Token id of the query token
        :rtype: int | None
        """
        if f"[VIE{viewer_token_id}]" in self.added_tokens_encoder:
            return self.added_tokens_encoder[f"[VIE{viewer_token_id}]"]
        return None



"""
Class factory module for Lightning IR.

This module provides factory classes for creating various components of the Lightning IR library
by extending Hugging Face Transformers classes.
"""

from __future__ import annotations

import importlib
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any

from transformers import (
    CONFIG_MAPPING,
    MODEL_MAPPING,
    TOKENIZER_MAPPING,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)
from transformers.models.auto.configuration_auto import model_type_to_module_name
from transformers.models.auto.tokenization_auto import TOKENIZER_MAPPING_NAMES, get_tokenizer_config

if TYPE_CHECKING:
    from . import LightningIRConfig, LightningIRModel, LightningIRTokenizer


def _get_model_class(config: PretrainedConfig | type[PretrainedConfig]) -> type[PreTrainedModel]:
    # https://github.com/huggingface/transformers/blob/356b3cd71d7bfb51c88fea3e8a0c054f3a457ab9/src/transformers/models/auto/auto_factory.py#L387
    if isinstance(config, type):
        supported_models = MODEL_MAPPING[config]
    else:
        supported_models = MODEL_MAPPING[type(config)]
    if not isinstance(supported_models, (list, tuple)):
        return supported_models

    if isinstance(config, type):
        # we cannot parse architectures from a config class, we need an instance for this
        return supported_models[0]

    name_to_model = {model.__name__: model for model in supported_models}
    architectures = getattr(config, "architectures", [])
    for arch in architectures:
        if arch in name_to_model:
            return name_to_model[arch]
        elif f"TF{arch}" in name_to_model:
            return name_to_model[f"TF{arch}"]
        elif f"Flax{arch}" in name_to_model:
            return name_to_model[f"Flax{arch}"]

    # If not architecture is set in the config or match the supported models, the first element of the tuple is the
    # defaults.
    return supported_models[0]


class LightningIRClassFactory(ABC):
    """Base class for creating derived Lightning IR classes from HuggingFace classes."""

    def __init__(self, MixinConfig: type[LightningIRConfig]) -> None:
        """Creates a new LightningIRClassFactory.

        Args:
            MixinConfig (type[LightningIRConfig]): LightningIRConfig mixin class.
        """
        if getattr(MixinConfig, "backbone_model_type", None) is not None:
            MixinConfig = MixinConfig.__bases__[0]
        self.MixinConfig = MixinConfig

    @staticmethod
    def get_backbone_config(model_name_or_path: str | Path) -> PretrainedConfig:
        """Grabs the configuration from a checkpoint of a pretrained HuggingFace model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            PretrainedConfig: Configuration of the backbone model.
        """
        backbone_model_type = LightningIRClassFactory.get_backbone_model_type(model_name_or_path)
        return CONFIG_MAPPING[backbone_model_type].from_pretrained(model_name_or_path)

    @staticmethod
    def get_lightning_ir_config(model_name_or_path: str | Path) -> LightningIRConfig | None:
        """Grabs the Lightning IR configuration from a checkpoint of a pretrained Lightning IR model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            LightningIRConfig | None: Configuration class of the Lightning IR model.
        """
        model_type = LightningIRClassFactory.get_lightning_ir_model_type(model_name_or_path)
        if model_type is None:
            return None
        return CONFIG_MAPPING[model_type].from_pretrained(model_name_or_path)

    @staticmethod
    def get_backbone_model_type(model_name_or_path: str | Path, *args, **kwargs) -> str:
        """Grabs the model type from a checkpoint of a pretrained HuggingFace model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            str: Model type of the backbone model.
        Raises:
            ValueError: If the type of the model is None in the configuration.
        """
        config_dict, _ = PretrainedConfig.get_config_dict(model_name_or_path, *args, **kwargs)
        backbone_model_type = config_dict.get("backbone_model_type", None) or config_dict.get("model_type")
        if backbone_model_type is None:
            raise ValueError(f"Unable to load PretrainedConfig from {model_name_or_path}")
        return backbone_model_type

    @staticmethod
    def get_lightning_ir_model_type(model_name_or_path: str | Path) -> str | None:
        """Grabs the Lightning IR model type from a checkpoint of a pretrained HuggingFace model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            str | None: Model type of the Lightning IR model.
        Raises:
            ValueError: If the backbone model type is not found in the configuration.
        """
        config_dict, _ = PretrainedConfig.get_config_dict(model_name_or_path)
        if "backbone_model_type" not in config_dict:
            return None
        return config_dict.get("model_type", None)

    def cc_lir_model_type(self, Config: type[LightningIRConfig]) -> str:
        """Camel case model type of the Lightning IR model."""
        return "".join(s.title() for s in Config.model_type.split("-"))

    @abstractmethod
    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> Any:
        """Loads a derived Lightning IR class from a pretrained HuggingFace model. Must be implemented by subclasses.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            Any: Derived Lightning IR class.
        """
        ...

    @abstractmethod
    def from_backbone_class(self, BackboneClass: type) -> type:
        """Creates a derived Lightning IR class from a backbone HuggingFace class. Must be implemented by subclasses.

        Args:
            BackboneClass (type): Backbone class.
        Returns:
            type: Derived Lightning IR class.
        """
        ...


class LightningIRConfigClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRConfig classes from HuggingFace configuration classes."""

    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> type[LightningIRConfig]:
        """Loads a derived LightningIRConfig from a pretrained HuggingFace model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            type[LightningIRConfig]: Derived LightningIRConfig.
        """
        backbone_config = self.get_backbone_config(model_name_or_path)
        DerivedLightningIRConfig = self.from_backbone_class(type(backbone_config))
        return DerivedLightningIRConfig

    def from_backbone_class(self, BackboneClass: type[PretrainedConfig]) -> type[LightningIRConfig]:
        """Creates a derived LightningIRConfig from a transformers.PretrainedConfig_ backbone configuration class. If
        the backbone configuration class is already a derived LightningIRConfig, it is returned as is.

        .. _transformers.PretrainedConfig: \
https://huggingface.co/docs/transformers/main_classes/configuration#transformers.PretrainedConfig

        Args:
            BackboneClass (type[PretrainedConfig]): Backbone configuration class.
        Returns:
            type[LightningIRConfig]: Derived LightningIRConfig.
        """
        if getattr(BackboneClass, "backbone_model_type", None) is not None:
            return BackboneClass
        LightningIRConfigMixin: type[LightningIRConfig] = CONFIG_MAPPING[self.MixinConfig.model_type]

        DerivedLightningIRConfig = type(
            f"{self.cc_lir_model_type(LightningIRConfigMixin)}{BackboneClass.__name__}",
            (LightningIRConfigMixin, BackboneClass),
            {
                "model_type": self.MixinConfig.model_type,
                "backbone_model_type": BackboneClass.model_type,
                "mixin_config": self.MixinConfig,
            },
        )
        return DerivedLightningIRConfig


class LightningIRModelClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRModel classes from HuggingFace model classes."""

    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> type[LightningIRModel]:
        """Loads a derived LightningIRModel from a pretrained HuggingFace model.

        Args:
            model_name_or_path (str | Path): Path to the model or its name.
        Returns:
            type[LightningIRModel]: Derived LightningIRModel.
        """
        backbone_config = self.get_backbone_config(model_name_or_path)
        BackboneModel = _get_model_class(backbone_config)
        DerivedLightningIRModel = self.from_backbone_class(BackboneModel)
        return DerivedLightningIRModel

    def from_backbone_class(self, BackboneClass: type[PreTrainedModel]) -> type[LightningIRModel]:
        """Creates a derived LightningIRModel from a transformers.PreTrainedModel_ backbone model. If the backbone model
          is already a LightningIRModel, it is returned as is.

        .. _transformers.PreTrainedModel: \
https://huggingface.co/transformers/main_classes/model#transformers.PreTrainedModel

        Args:
            BackboneClass (type[PreTrainedModel]): Backbone model class.
        Returns:
            type[LightningIRModel]: Derived LightningIRModel.
        Raises:
            ValueError: If the backbone model is not a valid backbone model.
            ValueError: If the backbone model is not a LightningIRModel and no LightningIRConfig is passed.
            ValueError: If the LightningIRModel mixin is not registered with the Hugging Face model mapping.
        """
        if getattr(BackboneClass.config_class, "backbone_model_type", None) is not None:
            return BackboneClass
        BackboneConfig = BackboneClass.config_class
        if BackboneConfig is None:
            raise ValueError(
                f"Model {BackboneClass} is not a valid backbone model because it is missing a `config_class`."
            )

        LightningIRModelMixin: type[LightningIRModel] = _get_model_class(self.MixinConfig)

        DerivedLightningIRConfig = LightningIRConfigClassFactory(self.MixinConfig).from_backbone_class(BackboneConfig)

        DerivedLightningIRModel = type(
            f"{self.cc_lir_model_type(LightningIRModelMixin.config_class)}{BackboneClass.__name__}",
            (LightningIRModelMixin, BackboneClass),
            {"config_class": DerivedLightningIRConfig, "_backbone_forward": BackboneClass.forward},
        )
        return DerivedLightningIRModel


class LightningIRTokenizerClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRTokenizer classes from HuggingFace tokenizer classes."""

    @staticmethod
    def get_backbone_config(model_name_or_path: str | Path) -> PretrainedConfig:
        """Grabs the tokenizer configuration class from a checkpoint of a pretrained HuggingFace tokenizer.

        Args:
            model_name_or_path (str | Path): Path to the tokenizer or its name.
        Returns:
            PretrainedConfig: Configuration class of the backbone tokenizer.
        """
        backbone_model_type = LightningIRTokenizerClassFactory.get_backbone_model_type(model_name_or_path)
        return CONFIG_MAPPING[backbone_model_type].from_pretrained(model_name_or_path)

    @staticmethod
    def get_backbone_model_type(model_name_or_path: str | Path, *args, **kwargs) -> str:
        """Grabs the model type from a checkpoint of a pretrained HuggingFace tokenizer.

        Args:
            model_name_or_path (str | Path): Path to the tokenizer or its name.
        Returns:
            str: Model type of the backbone tokenizer.
        """
        try:
            return LightningIRClassFactory.get_backbone_model_type(model_name_or_path, *args, **kwargs)
        except (OSError, ValueError) as err:
            # best guess at model type
            config_dict = get_tokenizer_config(model_name_or_path)
            backbone_tokenizer_class = config_dict.get("backbone_tokenizer_class", None)
            if backbone_tokenizer_class is not None:
                config_class = backbone_tokenizer_class.removesuffix("Fast").replace("Tokenizer", "Config")
                for module_name, tokenizers in TOKENIZER_MAPPING_NAMES.items():
                    if backbone_tokenizer_class in tokenizers:
                        module_name = model_type_to_module_name(module_name)
                        module = importlib.import_module(f".{module_name}", "transformers.models")
                        if hasattr(module, backbone_tokenizer_class) and hasattr(module, config_class):
                            config = getattr(module, config_class)
                            return config.model_type
            raise ValueError("No backbone model found in the configuration") from err

    def from_pretrained(
        self, model_name_or_path: str | Path, *args, use_fast: bool = True, **kwargs
    ) -> type[LightningIRTokenizer]:
        """Loads a derived LightningIRTokenizer from a pretrained HuggingFace tokenizer.

        Args:
            model_name_or_path (str | Path): Path to the tokenizer or its name.
            use_fast (bool, optional): Whether to use the fast tokenizer. Defaults to True.
        Returns:
            type[LightningIRTokenizer]: Derived LightningIRTokenizer.
        Raises:
            ValueError: If no fast tokenizer is found when `use_fast` is True.
            ValueError: If no slow tokenizer is found when `use_fast` is False.
        """
        backbone_config = self.get_backbone_config(model_name_or_path)
        BackboneTokenizers = TOKENIZER_MAPPING[type(backbone_config)]
        DerivedLightningIRTokenizers = self.from_backbone_classes(BackboneTokenizers, type(backbone_config))
        if use_fast:
            DerivedLightningIRTokenizer = DerivedLightningIRTokenizers[1]
            if DerivedLightningIRTokenizer is None:
                raise ValueError("No fast tokenizer found.")
        else:
            DerivedLightningIRTokenizer = DerivedLightningIRTokenizers[0]
            if DerivedLightningIRTokenizer is None:
                raise ValueError("No slow tokenizer found.")
        return DerivedLightningIRTokenizer

    def from_backbone_classes(
        self,
        BackboneClasses: tuple[type[PreTrainedTokenizerBase] | None, type[PreTrainedTokenizerBase] | None],
        BackboneConfig: type[PretrainedConfig] | None = None,
    ) -> tuple[type[LightningIRTokenizer] | None, type[LightningIRTokenizer] | None]:
        """Creates derived slow and fastLightningIRTokenizers from a tuple of backbone HuggingFace tokenizer classes.

        Args:
            BackboneClasses (tuple[type[PreTrainedTokenizerBase] | None, type[PreTrainedTokenizerBase] | None]):
                Slow and fast backbone tokenizer classes.
            BackboneConfig (type[PretrainedConfig] | None, optional): Backbone configuration class. Defaults to None.
        Returns:
            tuple[type[LightningIRTokenizer] | None, type[LightningIRTokenizer] | None]: Slow and fast derived
            LightningIRTokenizers.
        """
        DerivedLightningIRTokenizers = tuple(
            None if BackboneClass is None else self.from_backbone_class(BackboneClass)
            for BackboneClass in BackboneClasses
        )
        if DerivedLightningIRTokenizers[1] is not None:
            DerivedLightningIRTokenizers[1].slow_tokenizer_class = DerivedLightningIRTokenizers[0]
        return DerivedLightningIRTokenizers

    def from_backbone_class(self, BackboneClass: type[PreTrainedTokenizerBase]) -> type[LightningIRTokenizer]:
        """Creates a derived LightningIRTokenizer from a transformers.PreTrainedTokenizerBase_ backbone tokenizer. If
        the backbone tokenizer is already a LightningIRTokenizer, it is returned as is.

        .. _transformers.PreTrainedTokenizerBase: \
https://huggingface.co/transformers/main_classes/tokenizer.html#transformers.PreTrainedTokenizerBase

        Args:
            BackboneClass (type[PreTrainedTokenizerBase]): Backbone tokenizer class.
        Returns:
            type[LightningIRTokenizer]: Derived LightningIRTokenizer.
        """
        if hasattr(BackboneClass, "config_class"):
            return BackboneClass
        LightningIRTokenizerMixin = TOKENIZER_MAPPING[self.MixinConfig][0]

        DerivedLightningIRTokenizer = type(
            f"{self.cc_lir_model_type(LightningIRTokenizerMixin.config_class)}{BackboneClass.__name__}",
            (LightningIRTokenizerMixin, BackboneClass),
            {},
        )

        return DerivedLightningIRTokenizer



"""
Model hub for loading external models.

This module contains mappings and callbacks for external model checkpoints used in the Lightning IR library.

Attributes:
    CHECKPOINT_MAPPING (dict[str, LightningIRConfig]): Mapping of model checkpoint identifiers to their configurations.
    BACKBONE_MAPPING (dict[str, type]): Mapping of model checkpoint identifiers to their backbone model classes.
    STATE_DICT_KEY_MAPPING (dict[str, list[tuple[str | None, str]]]): Mapping of state dictionary keys for model
        checkpoints.
    POST_LOAD_CALLBACKS (dict[str, Callable[[LightningIRModel], LightningIRModel]]): Callbacks to be executed after
        loading a model.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from .config import LightningIRConfig
    from .model import LightningIRModel

CHECKPOINT_MAPPING: dict[str, LightningIRConfig] = {}
BACKBONE_MAPPING: dict[str, type] = {}
STATE_DICT_KEY_MAPPING: dict[str, list[tuple[str | None, str]]] = {}
POST_LOAD_CALLBACKS: dict[str, Callable[[LightningIRModel], LightningIRModel]] = {}



"""
Module module for cross-encoder models.

This module defines the Lightning IR module class used to implement cross-encoder models.
"""

from collections.abc import Mapping, Sequence
from typing import Any

import torch
from transformers import PreTrainedModel

from ..base.module import LightningIRModule
from ..data import RankBatch, SearchBatch, TrainBatch
from ..loss.base import LossFunction, ScoringLossFunction
from .cross_encoder_config import CrossEncoderConfig
from .cross_encoder_model import CrossEncoderModel, CrossEncoderOutput
from .cross_encoder_tokenizer import CrossEncoderTokenizer


class CrossEncoderModule(LightningIRModule):
    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: CrossEncoderConfig | None = None,
        model: CrossEncoderModel | None = None,
        BackboneModel: type[PreTrainedModel] | None = None,
        loss_functions: Sequence[LossFunction | tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
        model_kwargs: Mapping[str, Any] | None = None,
    ):
        """:class:`.LightningIRModule` for cross-encoder models. It contains a :class:`.CrossEncoderModel` and a
        :class:`.CrossEncoderTokenizer` and implements the training, validation, and testing steps for the model.

        .. _ir-measures: https://ir-measur.es/en/latest/index.html

        Args:
            model_name_or_path (str | None): Name or path of backbone model or fine-tuned Lightning IR model.
                Defaults to None.
            config (CrossEncoderConfig | None): CrossEncoderConfig to apply when loading from backbone model.
                Defaults to None.
            model (CrossEncoderModel | None): Already instantiated CrossEncoderModel. Defaults to None.
            BackboneModel (type[PreTrainedModel] | None): Huggingface PreTrainedModel class to use as backbone
                instead of the default AutoModel. Defaults to None.
            loss_functions (Sequence[LossFunction | tuple[LossFunction, float]] | None):
                Loss functions to apply during fine-tuning, optional loss weights can be provided per loss function.
                Defaults to None.
            evaluation_metrics (Sequence[str] | None): Metrics corresponding to ir-measures_ measure strings to apply
                during validation or testing. Defaults to None.
            model_kwargs (Mapping[str, Any] | None): Additional keyword arguments to pass to `from_pretrained` when
                loading a model. Defaults to None.
        """
        super().__init__(
            model_name_or_path=model_name_or_path,
            config=config,
            model=model,
            BackboneModel=BackboneModel,
            loss_functions=loss_functions,
            evaluation_metrics=evaluation_metrics,
            model_kwargs=model_kwargs,
        )
        self.model: CrossEncoderModel
        self.config: CrossEncoderConfig
        self.tokenizer: CrossEncoderTokenizer

    def forward(self, batch: RankBatch | TrainBatch | SearchBatch) -> CrossEncoderOutput:
        """Runs a forward pass of the model on a batch of data and returns the contextualized embeddings from the
        backbone model as well as the relevance scores.

        Args:
            batch (RankBatch | TrainBatch | SearchBatch): Batch of data to run the forward pass on.
        Returns:
            CrossEncoderOutput: Output of the model.
        Raises:
            ValueError: If the batch is a SearchBatch.
        """
        if isinstance(batch, SearchBatch):
            raise ValueError("Searching is not available for cross-encoders")
        queries = batch.queries
        docs = [d for docs in batch.docs for d in docs]
        num_docs = [len(docs) for docs in batch.docs]
        encoding = self.prepare_input(queries, docs, num_docs)
        output = self.model.forward(encoding["encoding"])
        return output

    def _compute_losses(self, batch: TrainBatch, output: CrossEncoderOutput) -> list[torch.Tensor]:
        """Computes the losses for a training batch."""
        if self.loss_functions is None:
            raise ValueError("loss_functions must be set in the module")

        output.scores = output.scores.view(len(batch.query_ids), -1)
        batch.targets = batch.targets.view(*output.scores.shape, -1)

        losses = []
        for loss_function, _ in self.loss_functions:
            if not isinstance(loss_function, ScoringLossFunction):
                raise RuntimeError(f"Loss function {loss_function} is not a scoring loss function")
            losses.append(loss_function.compute_loss(output, batch))
        return losses



"""
Lightning IR module for native models.

This module provides the classes, configurations and tokenizer for various models in the Lightning IR framework.
"""

from .bi_encoders import (
    CoilConfig,
    CoilEmbedding,
    CoilModel,
    ColConfig,
    ColModel,
    ColTokenizer,
    DprConfig,
    DprModel,
    MvrConfig,
    MvrModel,
    MvrTokenizer,
    SpladeConfig,
    SpladeModel,
    SpladeTokenizer,
    UniCoilConfig,
    UniCoilModel,
)
from .cross_encoders import MonoConfig, MonoModel, SetEncoderConfig, SetEncoderModel, SetEncoderTokenizer

__all__ = [
    "CoilConfig",
    "CoilEmbedding",
    "CoilModel",
    "ColConfig",
    "ColModel",
    "ColTokenizer",
    "DprConfig",
    "DprModel",
    "MonoConfig",
    "MonoModel",
    "MvrConfig",
    "MvrModel",
    "MvrTokenizer",
    "SetEncoderConfig",
    "SetEncoderModel",
    "SetEncoderTokenizer",
    "SpladeConfig",
    "SpladeModel",
    "SpladeTokenizer",
    "UniCoilConfig",
    "UniCoilModel",
]



"""
Datasets for Lightning IR that data loading and sampling.

This module defines several datasets that handle loading and sampling data for training and inference.
"""

import csv
import warnings
from collections.abc import Iterator, Sequence
from itertools import islice
from pathlib import Path
from typing import Any, Literal, Self

import ir_datasets
import numpy as np
import pandas as pd
import torch
from ir_datasets.formats import GenericDoc, GenericDocPair
from torch.distributed import get_rank, get_world_size
from torch.utils.data import Dataset, IterableDataset, get_worker_info

from .data import DocSample, QuerySample, RankSample
from .external_datasets.ir_datasets_utils import ScoredDocTuple

RUN_HEADER = ["query_id", "q0", "doc_id", "rank", "score", "system"]


class _DummyIterableDataset(IterableDataset):
    """Dummy iterable dataset to use when all inference datasets are skipped."""

    def __iter__(self) -> Iterator:
        yield from iter([])


class IRDataset:
    _SKIP: bool = False
    """set to True to skip the dataset during inference."""

    def __init__(self, dataset: str) -> None:
        """Initializes a new IRDataset.

        Args:
            dataset (str): Dataset name.
        """
        super().__init__()
        self._dataset = dataset
        self._queries = None
        self._docs = None
        self._qrels = None

    @property
    def dataset(self) -> str:
        """Dataset name.

        Returns:
            str: Dataset name.
        """
        return self.DASHED_DATASET_MAP.get(self._dataset, self._dataset)

    @property
    def dataset_id(self) -> str:
        """Dataset id.

        Returns:
            str: Dataset id.
        """
        if self.ir_dataset is None:
            return self.dataset
        return self.ir_dataset.dataset_id()

    @property
    def docs_dataset_id(self) -> str:
        """ID of the dataset containing the documents.

        Returns:
            str: Document dataset id.
        """
        return ir_datasets.docs_parent_id(self.dataset_id)

    @property
    def dashed_docs_dataset_id(self) -> str:
        """Dataset id with dashes instead of slashes for the documents dataset.

        Returns:
            str: Document dataset id with dashes.
        """
        return self.docs_dataset_id.replace("/", "-")

    @property
    def ir_dataset(self) -> ir_datasets.Dataset | None:
        """Instance of ir_datasets.Dataset.

        Returns:
            ir_datasets.Dataset | None: Instance of ir_datasets.Dataset or None if the dataset is not found.
        """
        try:
            return ir_datasets.load(self.dataset)
        except KeyError:
            return None

    @property
    def DASHED_DATASET_MAP(self) -> dict[str, str]:
        """Map of dataset names with dashes to dataset names with slashes.

        Returns:
            dict[str, str]: Dataset map.
        """
        return {dataset.replace("/", "-"): dataset for dataset in ir_datasets.registry._registered}

    @property
    def queries(self) -> pd.Series:
        """Queries in the dataset.

        Returns:
            pd.Series: Queries.
        Raises:
            ValueError: If no queries are found in the dataset.
        """
        if self._queries is None:
            if self.ir_dataset is None:
                raise ValueError(f"Unable to find dataset {self.dataset} in ir-datasets")
            queries_iter = self.ir_dataset.queries_iter()
            self._queries = pd.Series(
                {query.query_id: query.default_text() for query in queries_iter},
                name="text",
            )
            self._queries.index.name = "query_id"
        return self._queries

    @property
    def docs(self) -> ir_datasets.indices.Docstore | dict[str, GenericDoc]:
        """Documents in the dataset.

        Returns:
            ir_datasets.indices.Docstore | dict[str, GenericDoc]: Documents.
        Raises:
            ValueError: If no documents are found in the dataset.
        """
        if self._docs is None:
            if self.ir_dataset is None:
                raise ValueError(f"Unable to find dataset {self.dataset} in ir-datasets")
            self._docs = self.ir_dataset.docs_store()
        return self._docs

    @property
    def qrels(self) -> pd.DataFrame | None:
        """Qrels in the dataset.

        Returns:
            pd.DataFrame | None: Qrels.
        """
        if self._qrels is not None:
            return self._qrels
        if self.ir_dataset is None or not self.ir_dataset.has_qrels():
            return None
        qrels = pd.DataFrame(self.ir_dataset.qrels_iter()).rename({"subtopic_id": "iteration"}, axis=1)
        if "iteration" not in qrels.columns:
            qrels["iteration"] = 0
        qrels = qrels.drop_duplicates(["query_id", "doc_id", "iteration"])
        qrels = qrels.set_index(["query_id", "doc_id", "iteration"]).unstack(level=-1)
        self._qrels = qrels
        return self._qrels

    def prepare_constituent(self, constituent: Literal["qrels", "queries", "docs", "scoreddocs", "docpairs"]) -> None:
        """Downloads the constituent of the dataset using ir_datasets if needed.

        Args:
            constituent (Literal["qrels", "queries", "docs", "scoreddocs", "docpairs"]): Constituent to download.
        """
        if self.ir_dataset is None:
            return
        if self.ir_dataset.has(constituent):
            if constituent == "docs" and hasattr(self.ir_dataset, "docs_store"):
                # build docs store if not already built
                docs_store = self.ir_dataset.docs_store()
                if not docs_store.built():
                    docs_store.build()
            else:
                # get first item to trigger download
                next(getattr(self.ir_dataset, f"{constituent}_iter")())


class _DataParallelIterableDataset(IterableDataset):
    # https://github.com/Lightning-AI/pytorch-lightning/issues/15734
    def __init__(self) -> None:
        super().__init__()
        # TODO add support for multi-gpu and multi-worker inference; currently
        # doesn't work
        worker_info = get_worker_info()
        num_workers = worker_info.num_workers if worker_info is not None else 1
        worker_id = worker_info.id if worker_info is not None else 0

        try:
            world_size = get_world_size()
            process_rank = get_rank()
        except (RuntimeError, ValueError):
            world_size = 1
            process_rank = 0

        self.num_replicas = num_workers * world_size
        self.rank = process_rank * num_workers + worker_id


class QueryDataset(IRDataset, _DataParallelIterableDataset):
    def __init__(self, query_dataset: str, num_queries: int | None = None) -> None:
        """Dataset containing queries.

        Args:
            query_dataset (str): Path to file containing queries or valid ir_datasets id.
            num_queries (int | None, optional): Number of queries in dataset. If None, the number of queries will
                attempted to be inferred. Defaults to None.
        """
        super().__init__(query_dataset)
        super(IRDataset, self).__init__()
        self.num_queries = num_queries

    def __len__(self) -> int | None:
        """Number of queries in the dataset. Returns None if the number of queries cannot be inferred.

        Returns:
            int | None: Number of queries.
        """
        # TODO fix len for multi-gpu and multi-worker inference
        return self.num_queries or getattr(self.ir_dataset, "queries_count", lambda: None)() or None

    def __iter__(self) -> Iterator[QuerySample]:
        """Iterate over queries in the dataset.

        Yields:
            QuerySample: Query sample.
        """
        start = self.rank
        stop = self.num_queries
        step = self.num_replicas
        for sample in islice(self.ir_dataset.queries_iter(), start, stop, step):
            query_sample = QuerySample.from_ir_dataset_sample(sample)
            if self.qrels is not None:
                qrels = (
                    self.qrels.loc[[query_sample.query_id]]
                    .stack(future_stack=True)
                    .dropna()
                    .astype(int)
                    .reset_index()
                    .to_dict(orient="records")
                )
                query_sample.qrels = qrels
            yield query_sample

    def prepare_data(self) -> None:
        """Downloads queries using ir_datasets if needed."""
        self.prepare_constituent("queries")


class DocDataset(IRDataset, _DataParallelIterableDataset):
    def __init__(self, doc_dataset: str, num_docs: int | None = None, text_fields: Sequence[str] | None = None) -> None:
        """Dataset containing documents.

        Args:
            doc_dataset (str): Path to file containing documents or valid ir_datasets id.
            num_docs (int | None, optional): Number of documents in dataset. If None, the number of documents will
                attempted to be inferred. Defaults to None.
            text_fields (Sequence[str] | None, optional): Fields to parse the document text from. Defaults to None.
        """
        super().__init__(doc_dataset)
        super(IRDataset, self).__init__()
        self.num_docs = num_docs
        self.text_fields = text_fields

    def __len__(self) -> int | None:
        """Number of documents in the dataset. Returns None if the number of documents cannot be inferred.

        Returns:
            int | None: Number of documents.
        """
        # TODO fix len for multi-gpu and multi-worker inference
        return self.num_docs or getattr(self.ir_dataset, "docs_count", lambda: None)() or None

    def __iter__(self) -> Iterator[DocSample]:
        """Iterate over documents in the dataset.

        Yields:
            DocSample: Document sample.
        """
        start = self.rank
        stop = self.num_docs
        step = self.num_replicas
        for sample in islice(self.ir_dataset.docs_iter(), start, stop, step):
            yield DocSample.from_ir_dataset_sample(sample, self.text_fields)

    def prepare_data(self) -> None:
        """Downloads documents using ir_datasets if needed."""
        self.prepare_constituent("docs")


class Sampler:
    """Helper class for sampling subsets of documents from a ranked list."""

    @staticmethod
    def single_relevant(documents: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        """Sampling strategy to randomly sample a single relevant document. The remaining ``sample_size - 1``
        are non-relevant.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        relevance = documents.filter(like="relevance").max(axis=1).fillna(0)
        relevant = documents.loc[relevance.gt(0)].sample(1)
        non_relevant_bool = relevance.eq(0) & ~documents["rank"].isna()
        num_non_relevant = non_relevant_bool.sum()
        sample_non_relevant = min(sample_size - 1, num_non_relevant)
        non_relevant = documents.loc[non_relevant_bool].sample(sample_non_relevant)
        return pd.concat([relevant, non_relevant])

    @staticmethod
    def top(documents: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        """Sampling strategy to randomly sample a single relevant document. The remaining ``sample_size - 1``
        are non-relevant.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        return documents.head(sample_size)

    @staticmethod
    def top_and_random(documents: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        """Sampling strategy to randomly sample half the ``sample_size`` documents from the top of the ranking and the
        other half randomly.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        top_size = sample_size // 2
        random_size = sample_size - top_size
        top = documents.head(top_size)
        random = documents.iloc[top_size:].sample(random_size)
        return pd.concat([top, random])

    @staticmethod
    def random(documents: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        """Sampling strategy to randomly sample ``sample_size`` documents.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        return documents.sample(sample_size)

    @staticmethod
    def log_random(documents: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        """Sampling strategy to randomly sample documents with a higher probability to sample documents from the top of
        the ranking.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        weights = 1 / np.log1p(documents["rank"])
        weights[weights.isna()] = weights.min()
        return documents.sample(sample_size, weights=weights)

    @staticmethod
    def sample(
        df: pd.DataFrame,
        sample_size: int,
        sampling_strategy: Literal["single_relevant", "top", "random", "log_random", "top_and_random"],
    ) -> pd.DataFrame:
        """
        Samples a subset of documents from a ranked list given a sampling_strategy.

        Args:
            documents (pd.DataFrame): Ranked list of documents.
            sample_size (int): Number of documents to sample.
        Returns:
            pd.DataFrame: Sampled documents.
        """
        if sample_size == -1:
            return df
        if hasattr(Sampler, sampling_strategy):
            return getattr(Sampler, sampling_strategy)(df, sample_size)
        raise ValueError("Invalid sampling strategy.")


class RunDataset(IRDataset, Dataset):
    def __init__(
        self,
        run_path_or_id: Path | str,
        depth: int = -1,
        sample_size: int = -1,
        sampling_strategy: Literal["single_relevant", "top", "random", "log_random", "top_and_random"] = "top",
        targets: Literal["relevance", "subtopic_relevance", "rank", "score"] | None = None,
        normalize_targets: bool = False,
        add_docs_not_in_ranking: bool = False,
    ) -> None:
        """Dataset containing a list of queries with a ranked list of documents per query. Subsets of the ranked list
        can be sampled using different sampling strategies.

        Args:
            run_path_or_id (Path | str): Path to a run file or valid ir_datasets id.
            depth (int): Depth at which to cut off the ranking. If -1 the full ranking is kept.
                Defaults to -1.
            sample_size (int): The number of documents to sample per query. Defaults to -1.
            sampling_strategy (Literal["single_relevant", "top", "random", "log_random", "top_and_random"]):
                The sample strategy to use to sample documents. Defaults to "top".
            targets (Literal["relevance", "subtopic_relevance", "rank", "score"] | None):
                The data type to use as targets for a model during fine-tuning. If "relevance" the relevance
                judgements are parsed from qrels. Defaults to None.
            normalize_targets (bool): Whether to normalize the targets between 0 and 1. Defaults to False.
            add_docs_not_in_ranking (bool): Whether to add relevant documents to a sample that are in the qrels but not
                in the ranking. Defaults to False.
        """
        self.run_path = None
        if Path(run_path_or_id).is_file():
            self.run_path = Path(run_path_or_id)
            dataset = self.run_path.name.split(".")[0].split("__")[-1]
        else:
            dataset = str(run_path_or_id)
        super().__init__(dataset)
        self.depth = depth
        self.sample_size = sample_size
        self.sampling_strategy = sampling_strategy
        self.targets = targets
        self.normalize_targets = normalize_targets
        self.add_docs_not_in_ranking = add_docs_not_in_ranking

        if self.sampling_strategy == "top" and self.sample_size > self.depth:
            warnings.warn(
                "Sample size is greater than depth and top sampling strategy is used. "
                "This can cause documents to be sampled that are not contained "
                "in the run file, but that are present in the qrels.",
                stacklevel=2,
            )

        self.run: pd.DataFrame | None = None

    def prepare_data(self) -> Self:
        """Downloads docs, queries, scoreddocs, and qrels using ir_datasets if needed and available."""
        self.prepare_constituent("docs")
        self.prepare_constituent("queries")
        if self.run_path is None:
            self.prepare_constituent("scoreddocs")
            self.prepare_constituent("docpairs")
        self.prepare_constituent("qrels")
        return self

    def _setup(self):
        if self.run is not None:
            return
        self.run = self._load_run()
        self.run = self.run.drop_duplicates(["query_id", "doc_id"])

        if self.qrels is not None:
            run_query_ids = pd.Index(self.run["query_id"].drop_duplicates())
            qrels_query_ids = self.qrels.index.get_level_values("query_id").unique()
            query_ids = run_query_ids.intersection(qrels_query_ids)
            if len(run_query_ids.difference(qrels_query_ids)):
                self.run = self.run[self.run["query_id"].isin(query_ids)]
            # outer join if docs are from ir_datasets else only keep docs in run
            how = "left"
            if self._docs is None and self.add_docs_not_in_ranking:
                how = "outer"
            self.run = self.run.merge(
                self.qrels.loc[pd.IndexSlice[query_ids, :]].droplevel(0, axis=1).add_prefix("relevance_", axis=1),
                on=["query_id", "doc_id"],
                how=how,
            )

        if self.sample_size != -1:
            num_docs_per_query = self.run.groupby("query_id").transform("size")
            self.run = self.run[num_docs_per_query >= self.sample_size]

        self.run = self.run.sort_values(["query_id", "rank"])
        self.run_groups = self.run.groupby("query_id")
        self.query_ids = list(self.run_groups.groups.keys())

        if self.depth != -1 and self.run["rank"].max() < self.depth:
            warnings.warn("Depth is greater than the maximum rank in the run file.", stacklevel=2)

    @staticmethod
    def _load_csv(path: Path) -> pd.DataFrame:
        return pd.read_csv(
            path,
            sep=r"\s+",
            header=None,
            names=RUN_HEADER,
            usecols=[0, 1, 2, 3, 4],
            dtype={"query_id": str, "doc_id": str},
            quoting=csv.QUOTE_NONE,
            na_filter=False,
        )

    @staticmethod
    def _load_parquet(path: Path) -> pd.DataFrame:
        return pd.read_parquet(path)

    @staticmethod
    def _load_json(path: Path) -> pd.DataFrame:
        kwargs: dict[str, Any] = {}
        if ".jsonl" in path.suffixes:
            kwargs["lines"] = True
            kwargs["orient"] = "records"
        run = pd.read_json(path, **kwargs)
        return run

    def _get_run_path(self) -> Path | None:
        run_path = self.run_path
        if run_path is None:
            if self.ir_dataset is None or not self.ir_dataset.has_scoreddocs():
                raise ValueError(f"Run file or dataset with scoreddocs required. Got {self._dataset}")
            try:
                run_path = self.ir_dataset.scoreddocs_handler().scoreddocs_path()
            except NotImplementedError:
                pass
        return run_path

    def _clean_run(self, run: pd.DataFrame) -> pd.DataFrame:
        run = run.rename(
            {"qid": "query_id", "docid": "doc_id", "docno": "doc_id", "Q0": "iteration", "q0": "iteration"},
            axis=1,
        )
        dtypes = {"rank": np.int32, "query_id": str, "doc_id": str}
        if "score" in run.columns:
            dtypes["score"] = np.float32
        run = run.astype(dtypes)
        if "query" in run.columns:
            self._queries = run.drop_duplicates("query_id").set_index("query_id")["query"].rename("text")
            run = run.drop("query", axis=1)
        if "text" in run.columns:
            self._docs = (
                run.drop_duplicates("doc_id").set_index("doc_id")["text"].map(lambda x: GenericDoc("", x)).to_dict()
            )
            run = run.drop("text", axis=1)
        if self.depth != -1:
            run = run[run["rank"] <= self.depth]
        return run

    def _load_run(self) -> pd.DataFrame:
        suffix_load_map = {
            ".tsv": self._load_csv,
            ".run": self._load_csv,
            ".csv": self._load_csv,
            ".parquet": self._load_parquet,
            ".json": self._load_json,
            ".jsonl": self._load_json,
        }
        run = None

        # try loading run from file
        run_path = self._get_run_path()
        if run_path is not None:
            load_func = suffix_load_map.get(run_path.suffixes[0], None)
            if load_func is not None:
                try:
                    run = load_func(run_path)
                except Exception:
                    pass

        # try loading run from ir_datasets
        if run is None and self.ir_dataset is not None and self.ir_dataset.has_scoreddocs():
            run = pd.DataFrame(self.ir_dataset.scoreddocs_iter())
            run["rank"] = run.groupby("query_id")["score"].rank("first", ascending=False)
            run = run.sort_values(["query_id", "rank"])

        if run is None:
            raise ValueError("Invalid run file format.")

        run = self._clean_run(run)
        return run

    @property
    def qrels(self) -> pd.DataFrame | None:
        """The qrels in the dataset. If the dataset does not contain qrels, the qrels are None.

        Returns:
            pd.DataFrame | None: Qrels.
        """
        if self._qrels is not None:
            return self._qrels
        if self.run is not None and "relevance" in self.run:
            qrels = self.run[["query_id", "doc_id", "relevance"]].copy()
            if "iteration" in self.run:
                qrels["iteration"] = self.run["iteration"]
            else:
                qrels["iteration"] = "0"
            self.run = self.run.drop(["relevance", "iteration"], axis=1, errors="ignore")
            qrels = qrels.drop_duplicates(["query_id", "doc_id", "iteration"])
            qrels = qrels.set_index(["query_id", "doc_id", "iteration"]).unstack(level=-1)
            self._qrels = qrels
            return self._qrels
        return super().qrels

    def __len__(self) -> int:
        """Number of queries in the dataset.

        Returns:
            int: Number of queries.
        """
        self._setup()
        return len(self.query_ids)

    def __getitem__(self, idx: int) -> RankSample:
        """Samples a single query and corresponding ranked documents from the run. The documents are sampled according
        to the sampling strategy and sample size.

        Args:
            idx (int): Index of the query.
        Returns:
            RankSample: Sampled query and documents.
        Raises:
            ValueError: If the targets are not found in the run file.
        """
        self._setup()
        query_id = str(self.query_ids[idx])
        group = self.run_groups.get_group(query_id).copy()
        query = self.queries[query_id]
        group = Sampler.sample(group, self.sample_size, self.sampling_strategy)

        doc_ids = tuple(group["doc_id"])
        docs = tuple(self.docs.get(doc_id).default_text() for doc_id in doc_ids)

        targets = None
        if self.targets is not None:
            filtered = group.set_index("doc_id").loc[list(doc_ids)].filter(like=self.targets).fillna(0)
            if filtered.empty:
                raise ValueError(f"targets `{self.targets}` not found in run file")
            targets = torch.from_numpy(filtered.values)
            if self.targets == "rank":
                # invert ranks to be higher is better (necessary for loss functions)
                targets = self.depth - targets + 1
            if self.normalize_targets:
                targets_min = targets.min()
                targets_max = targets.max()
                targets = (targets - targets_min) / (targets_max - targets_min)
        qrels = None
        if self.qrels is not None:
            qrels = (
                self.qrels.loc[[query_id]]
                .stack(future_stack=True)
                .dropna()
                .astype(int)
                .reset_index()
                .to_dict(orient="records")
            )
        return RankSample(query_id, query, doc_ids, docs, targets, qrels)


class TupleDataset(IRDataset, IterableDataset):
    def __init__(
        self,
        tuples_dataset: str,
        targets: Literal["order", "score"] = "order",
        num_docs: int | None = None,
    ) -> None:
        """Dataset containing tuples of a query and n-documents. Used for fine-tuning models on ranking tasks.

        Args:
            tuples_dataset (str): Path to file containing tuples or valid ir_datasets id.
            targets (Literal["order", "score"], optional): Data type to use as targets for a model during fine-tuning.
                Defaults to "order".
            num_docs (int | None, optional): Maximum number of documents per query. Defaults to None.
        """
        super().__init__(tuples_dataset)
        super(IRDataset, self).__init__()
        self.targets = targets
        self.num_docs = num_docs

    def _parse_sample(
        self, sample: ScoredDocTuple | GenericDocPair
    ) -> tuple[tuple[str, ...], tuple[str, ...], tuple[float, ...] | None]:
        if isinstance(sample, GenericDocPair):
            if self.targets == "score":
                raise ValueError("ScoredDocTuple required for score targets.")
            targets = (1.0, 0.0)
            doc_ids = (sample.doc_id_a, sample.doc_id_b)
        elif isinstance(sample, ScoredDocTuple):
            doc_ids = sample.doc_ids[: self.num_docs]
            if self.targets == "score":
                if sample.scores is None:
                    raise ValueError("tuples dataset does not contain scores")
                targets = sample.scores
            elif self.targets == "order":
                targets = tuple([1.0] + [0.0] * (sample.num_docs - 1))
            else:
                raise ValueError(f"invalid value for targets, got {self.targets}, expected one of (order, score)")
            targets = targets[: self.num_docs]
        else:
            raise ValueError("Invalid sample type.")
        docs = tuple(self.docs.get(doc_id).default_text() for doc_id in doc_ids)
        return doc_ids, docs, targets

    def __iter__(self) -> Iterator[RankSample]:
        """Iterates over tuples in the dataset.

        Yields:
            RankSample: Sampled query and documents with targets.
        """
        for sample in self.ir_dataset.docpairs_iter():
            query_id = sample.query_id
            query = self.queries.loc[query_id]
            doc_ids, docs, targets = self._parse_sample(sample)
            if targets is not None:
                targets = torch.tensor(targets)
            yield RankSample(query_id, query, doc_ids, docs, targets)

    def prepare_data(self) -> None:
        """Downloads tuples using ir_datasets if needed."""
        self.prepare_constituent("docs")
        self.prepare_constituent("queries")
        self.prepare_constituent("docpairs")



"""
Configuration and model implementation for SetEncoder type models. Originally proposed in
`set-Encoder: Permutation-Invariant Inter-passage Attention for Listwise Passage Re-ranking with Cross-Encoders
<https://link.springer.com/chapter/10.1007/978-3-031-88711-6_1>`_.
"""

from collections.abc import Sequence
from functools import partial

import torch
from tokenizers.processors import TemplateProcessing
from transformers import BatchEncoding

from ...cross_encoder import CrossEncoderOutput, CrossEncoderTokenizer
from .mono import MonoConfig, MonoModel


class SetEncoderConfig(MonoConfig):
    """Configuration class for a SetEncoder model."""

    model_type = "set-encoder"
    """Model type for a SetEncoder model."""

    def __init__(
        self,
        *args,
        depth: int = 100,
        add_extra_token: bool = False,
        sample_missing_docs: bool = True,
        **kwargs,
    ):
        """
        A SetEncoder model encodes a query and a set of documents jointly.
        Each document's embedding is updated with context from the entire set,
        and a relevance score is computed per document using a linear layer.

        Args:
            depth (int): Number of documents to encode per query. Defaults to 100.
            add_extra_token (bool): Whether to add an extra token to the input sequence to separate
                the query from the documents. Defaults to False.
            sample_missing_docs (bool): Whether to sample missing documents when the number of documents is less
                than the specified depth. Defaults to True.
        """

        super().__init__(*args, **kwargs)
        self.depth = depth
        self.add_extra_token = add_extra_token
        self.sample_missing_docs = sample_missing_docs


class SetEncoderModel(MonoModel):
    """SetEncoder model. See :class:`SetEncoderConfig` for configuration options."""

    config_class = SetEncoderConfig
    self_attention_pattern = "self"

    ALLOW_SUB_BATCHING = False  # listwise model

    def __init__(self, config: SetEncoderConfig, *args, **kwargs):
        """Initializes a SetEncoder model give a :class:`SetEncoderConfig`.

        Args:
            config (SetEncoderConfig): Configuration for the SetEncoder model.
        """
        super().__init__(config, *args, **kwargs)
        self.config: SetEncoderConfig
        self.attn_implementation = "eager"
        if self.config.backbone_model_type is not None and self.config.backbone_model_type not in ("bert", "electra"):
            raise ValueError(
                f"SetEncoderModel does not support backbone model type {self.config.backbone_model_type}. "
                f"Supported types are 'bert' and 'electra'."
            )

    def get_extended_attention_mask(
        self,
        attention_mask: torch.Tensor,
        input_shape: tuple[int, ...],
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        num_docs: Sequence[int] | None = None,
    ) -> torch.Tensor:
        """
        Extends the attention mask to account for the number of documents per query.

        Args:
            attention_mask (torch.Tensor): Attention mask for the input sequence.
            input_shape (tuple[int, ...]): Shape of the input sequence.
            device (torch.device | None): Device to move the attention mask to. Defaults to None.
            dtype (torch.dtype | None): Data type of the attention mask. Defaults to None.
            num_docs (Sequence[int] | None): Specifies how many documents are passed per query. If a sequence of
                integers, `len(num_doc)` should be equal to the number of queries and `sum(num_docs)` equal to the
                number of documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries. Defaults to None.
            Returns:
                torch.Tensor: Extended attention mask.
        """
        if num_docs is not None:
            eye = (1 - torch.eye(self.config.depth, device=device)).long()
            if not self.config.sample_missing_docs:
                eye = eye[:, : max(num_docs)]
            other_doc_attention_mask = torch.cat([eye[:n] for n in num_docs])
            attention_mask = torch.cat(
                [attention_mask, other_doc_attention_mask.to(attention_mask)],
                dim=-1,
            )
            input_shape = tuple(attention_mask.shape)
        return super().get_extended_attention_mask(attention_mask, input_shape, device, dtype)

    def forward(self, encoding: BatchEncoding) -> CrossEncoderOutput:
        """Computes contextualized embeddings for the joint query-document input sequence and computes a relevance
        score.

        Args:
            encoding (BatchEncoding): Tokenizer encoding for the joint query-document input sequence.
        Returns:
            CrossEncoderOutput: Output of the model.
        """
        num_docs = encoding.pop("num_docs", None)
        self.get_extended_attention_mask = partial(self.get_extended_attention_mask, num_docs=num_docs)
        for name, module in self.named_modules():
            if name.endswith(self.self_attention_pattern):
                module.forward = partial(self.attention_forward, self, module, num_docs=num_docs)
        return super().forward(encoding)

    @staticmethod
    def attention_forward(
        _self,
        self: torch.nn.Module,
        hidden_states: torch.Tensor,
        attention_mask: torch.FloatTensor | None,
        *args,
        num_docs: Sequence[int],
        **kwargs,
    ) -> tuple[torch.Tensor]:
        """Performs the attention forward pass for the SetEncoder model.

        Args:
            _self (SetEncoderModel): Reference to the SetEncoder instance.
            self (torch.nn.Module): Reference to the attention module.
            hidden_states (torch.Tensor): Hidden states from the previous layer.
            attention_mask (torch.FloatTensor | None): Attention mask for the input sequence.
            num_docs (Sequence[int]): Specifies how many documents are passed per query. If a sequence of integers,
                `len(num_doc)` should be equal to the number of queries and `sum(num_docs)` equal to the number of
                documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries.
        Returns:
            tuple[torch.Tensor]: Contextualized embeddings.
        """
        key_value_hidden_states = hidden_states
        if num_docs is not None:
            key_value_hidden_states = _self.cat_other_doc_hidden_states(hidden_states, num_docs)

        batch_size = hidden_states.shape[0]
        query = (
            self.query(hidden_states)
            .view(batch_size, -1, self.num_attention_heads, self.attention_head_size)
            .transpose(1, 2)
        )
        key = (
            self.key(key_value_hidden_states)
            .view(batch_size, -1, self.num_attention_heads, self.attention_head_size)
            .transpose(1, 2)
        )
        value = (
            self.value(key_value_hidden_states)
            .view(batch_size, -1, self.num_attention_heads, self.attention_head_size)
            .transpose(1, 2)
        )

        context = torch.nn.functional.scaled_dot_product_attention(
            query,
            key,
            value,
            attention_mask.to(query.dtype) if attention_mask is not None else None,
            self.dropout.p if self.training else 0,
        )

        context = context.permute(0, 2, 1, 3).contiguous()
        new_context_shape = context.size()[:-2] + (self.all_head_size,)
        context = context.view(new_context_shape)
        return (context,)

    def cat_other_doc_hidden_states(
        self,
        hidden_states: torch.Tensor,
        num_docs: Sequence[int],
    ) -> torch.Tensor:
        """Concatenates the hidden states of other documents to the hidden states of the query and documents.

        Args:
            hidden_states (torch.Tensor): Hidden states of the query and documents.
            num_docs (Sequence[int]): Specifies how many documents are passed per query. If a sequence of integers,
                `len(num_doc)` should be equal to the number of queries and `sum(num_docs)` equal to the number of
                documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries.
        Returns:
            torch.Tensor: Concatenated hidden states of the query and documents.
        """
        idx = 1 if self.config.add_extra_token else 0
        split_other_doc_hidden_states = torch.split(hidden_states[:, idx], list(num_docs))
        repeated_other_doc_hidden_states = []
        for idx, h_states in enumerate(split_other_doc_hidden_states):
            missing_docs = 0 if self.config.depth is None else self.config.depth - num_docs[idx]
            if missing_docs and self.config.sample_missing_docs:
                mean = h_states.mean(0, keepdim=True).expand(missing_docs, -1)
                if num_docs[idx] == 1:
                    std = torch.zeros_like(mean)
                else:
                    std = h_states.std(0, keepdim=True).expand(missing_docs, -1)
                sampled_h_states = torch.normal(mean, std).to(h_states)
                h_states = torch.cat([h_states, sampled_h_states])
            repeated_other_doc_hidden_states.append(h_states.unsqueeze(0).expand(num_docs[idx], -1, -1))
        other_doc_hidden_states = torch.cat(repeated_other_doc_hidden_states)
        key_value_hidden_states = torch.cat([hidden_states, other_doc_hidden_states], dim=1)
        return key_value_hidden_states


class SetEncoderTokenizer(CrossEncoderTokenizer):
    config_class = SetEncoderConfig
    """Configuration class for the tokenizer."""

    def __init__(
        self,
        *args,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        add_extra_token: bool = False,
        **kwargs,
    ):
        """Initializes a SetEncoder tokenizer.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            add_extra_token (bool): Whether to add an extra interaction token. Defaults to False.
        """
        super().__init__(
            *args, query_length=query_length, doc_length=doc_length, add_extra_token=add_extra_token, **kwargs
        )
        self.add_extra_token = add_extra_token
        self.interaction_token = "[INT]"
        if add_extra_token:
            self.add_tokens([self.interaction_token], special_tokens=True)
            self._tokenizer.post_processor = TemplateProcessing(
                single="[CLS] $0 [SEP]",
                pair="[CLS] [INT] $A [SEP] $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", self.cls_token_id),
                    ("[SEP]", self.sep_token_id),
                    ("[INT]", self.interaction_token_id),
                ],
            )

    @property
    def interaction_token_id(self) -> int:
        if self.interaction_token in self.added_tokens_encoder:
            return self.added_tokens_encoder[self.interaction_token]
        raise ValueError(f"Token {self.interaction_token} not found in tokenizer")

    def tokenize(
        self,
        queries: str | Sequence[str] | None = None,
        docs: str | Sequence[str] | None = None,
        num_docs: Sequence[int] | int | None = None,
        **kwargs,
    ) -> dict[str, BatchEncoding]:
        """Tokenizes queries and documents into a single sequence of tokens.

        Args:
            queries (str | Sequence[str] | None): Queries to tokenize. Defaults to None.
            docs (str | Sequence[str] | None): Documents to tokenize. Defaults to None.
            num_docs (Sequence[int] | int | None): Specifies how many documents are passed per query. If a sequence of
                integers, `len(num_docs)` should be equal to the number of queries and `sum(num_docs)` equal to the
                number of documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries. Defaults to None.
            Returns:
                dict[str, BatchEncoding]: Tokenized query-document sequence.
            Raises:
                ValueError: If both queries and docs are None.
                ValueError: If queries and docs are not both lists or both strings.
        """
        encoding_dict = super().tokenize(queries, docs, num_docs, **kwargs)
        encoding_dict["encoding"] = BatchEncoding({**encoding_dict["encoding"], "num_docs": num_docs})
        return encoding_dict



"""Learning rate schedulers for LightningIR."""

import torch

from .schedulers import ConstantSchedulerWithLinearWarmup, LambdaWarmupScheduler, LinearSchedulerWithLinearWarmup


class WarmupLRScheduler(LambdaWarmupScheduler, torch.optim.lr_scheduler.LambdaLR):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        num_warmup_steps: int,
        **kwargs,
    ) -> None:
        """Base class for learning rate schedulers with warmup.

        Args:
            optimizer (torch.optim.Optimizer): Optimizer to adjust the learning rate for.
            num_warmup_steps (int): Number of warmup steps.
        """
        last_epoch = -1
        self.interval = "step"
        super().__init__(
            optimizer=optimizer,
            lr_lambda=self.value_lambda,
            num_warmup_steps=num_warmup_steps,
            last_epoch=last_epoch,
            **kwargs,
        )


class LinearLRSchedulerWithLinearWarmup(WarmupLRScheduler, LinearSchedulerWithLinearWarmup):
    """Scheduler for linearly decreasing learning rate with linear warmup."""

    pass


class ConstantLRSchedulerWithLinearWarmup(WarmupLRScheduler, ConstantSchedulerWithLinearWarmup):
    """Scheduler for constant learning rate with linear warmup."""

    pass


LR_SCHEDULERS = (
    LinearLRSchedulerWithLinearWarmup,
    ConstantLRSchedulerWithLinearWarmup,
    WarmupLRScheduler,
    torch.optim.lr_scheduler.LRScheduler,
)



from collections.abc import Sequence
from pathlib import Path

import ir_datasets
import pandas as pd
import pytest
from _pytest.fixtures import SubRequest

from lightning_ir import BiEncoderModule, LightningIRDataModule, LightningIRModule, LightningIRTrainer, RunDataset
from lightning_ir.callbacks import IndexCallback, RegisterLocalDatasetCallback, ReRankCallback, SearchCallback
from lightning_ir.models import DprConfig
from lightning_ir.retrieve import (
    FaissFlatIndexConfig,
    FaissIVFIndexConfig,
    FaissSearchConfig,
    IndexConfig,
    PlaidIndexConfig,
    PlaidSearchConfig,
    SearchConfig,
    SeismicIndexConfig,
    SeismicSearchConfig,
    TorchDenseIndexConfig,
    TorchDenseSearchConfig,
    TorchSparseIndexConfig,
    TorchSparseSearchConfig,
)

from .conftest import CORPUS_DIR, DATA_DIR


@pytest.fixture(
    params=[
        FaissFlatIndexConfig(),
        FaissIVFIndexConfig(num_centroids=16),
        TorchSparseIndexConfig(),
        TorchDenseIndexConfig(),
        PlaidIndexConfig(num_centroids=8),
        SeismicIndexConfig(num_postings=32),
    ],
    ids=["Faiss", "FaissIVF", "Sparse", "Dense", "Plaid", "Seismic"],
)
def index_config(request: SubRequest) -> IndexConfig:
    return request.param


def run_datamodule(module: LightningIRModule, inference_datasets: Sequence[RunDataset]) -> LightningIRDataModule:
    datamodule = LightningIRDataModule(num_workers=0, inference_batch_size=2, inference_datasets=inference_datasets)
    datamodule.setup(stage="test")
    return datamodule


# @pytest.mark.parametrize("devices", (1, 2))
def test_index_callback(
    tmp_path: Path,
    bi_encoder_module: BiEncoderModule,
    doc_datamodule: LightningIRDataModule,
    index_config: IndexConfig,
    # devices: int,
):
    if bi_encoder_module.config.model_type not in index_config.SUPPORTED_MODELS:
        pytest.skip(
            f"Indexing not supported for {bi_encoder_module.config.__class__.__name__} model "
            f"and {index_config.__class__.__name__} indexer"
        )

    index_dir = tmp_path / "index"
    index_callback = IndexCallback(index_config=index_config, index_dir=index_dir)

    trainer = LightningIRTrainer(
        # devices=devices,
        logger=False,
        enable_checkpointing=False,
        callbacks=[index_callback],
    )
    trainer.index(bi_encoder_module, datamodule=doc_datamodule)

    assert doc_datamodule.inference_datasets is not None
    assert index_callback.indexer.num_embeddings and index_callback.indexer.num_docs
    assert index_callback.indexer.num_embeddings >= index_callback.indexer.num_docs

    dataset_id = doc_datamodule.inference_datasets[0].dataset_id
    index_dir = index_dir / dataset_id
    assert (
        (index_dir / "index.faiss").exists()  # faiss
        or (index_dir / "index.pt").exists()  # sparse
        or (index_dir / "centroids.npy").exists()  # plaid
        or (index_dir / ".index.seismic").exists()  # seismic
    )
    assert (index_dir / "doc_ids.txt").exists()
    doc_ids_path = index_dir / "doc_ids.txt"
    doc_ids = doc_ids_path.read_text().split()
    for idx, doc_id in enumerate(doc_ids):
        assert doc_id == f"doc_id_{idx + 1}"
    assert (index_dir / "config.json").exists()


def get_index(
    bi_encoder_module: BiEncoderModule,
    doc_datamodule: LightningIRDataModule,
    search_config: SearchConfig,
) -> Path:
    index_config: IndexConfig
    if isinstance(search_config, FaissSearchConfig):
        index_type = "faiss"
        index_config = FaissFlatIndexConfig()
    elif isinstance(search_config, TorchSparseSearchConfig):
        index_type = "sparse"
        index_config = TorchSparseIndexConfig()
    elif isinstance(search_config, TorchDenseSearchConfig):
        index_type = "dense"
        index_config = TorchDenseIndexConfig()
    elif isinstance(search_config, PlaidSearchConfig):
        index_type = "plaid"
        index_config = PlaidIndexConfig(num_centroids=8)
    elif isinstance(search_config, SeismicSearchConfig):
        index_type = "seismic"
        index_config = SeismicIndexConfig(num_postings=32)
    else:
        raise ValueError("Unknown search_config type")
    index_dir = (
        DATA_DIR
        / "indexes"
        / f"{index_type}-{bi_encoder_module.config.model_type}-{bi_encoder_module.config.similarity_function}"
    )
    if index_dir.exists():
        return index_dir

    index_callback = IndexCallback(index_config=index_config, index_dir=index_dir)

    trainer = LightningIRTrainer(
        logger=False,
        enable_checkpointing=False,
        callbacks=[index_callback],
    )
    trainer.test(bi_encoder_module, datamodule=doc_datamodule)
    return index_dir


@pytest.mark.parametrize(
    "search_config",
    (
        FaissSearchConfig(k=3, imputation_strategy="min", candidate_k=3),
        FaissSearchConfig(k=3, imputation_strategy="gather", candidate_k=3),
        PlaidSearchConfig(k=3, centroid_score_threshold=0),
        TorchSparseSearchConfig(k=3),
        TorchDenseSearchConfig(k=3),
        SeismicSearchConfig(k=3),
    ),
    ids=["FaissMin", "FaissGather", "Plaid", "Sparse", "Dense", "Seismic"],
)
def test_search_callback(
    tmp_path: Path,
    bi_encoder_module: BiEncoderModule,
    query_datamodule: LightningIRDataModule,
    doc_datamodule: LightningIRDataModule,
    search_config: SearchConfig,
):
    if bi_encoder_module.config.model_type not in search_config.SUPPORTED_MODELS:
        pytest.skip(
            f"Searching not supported for {bi_encoder_module.config.__class__.__name__} model and "
            f"{search_config.__class__.__name__} searcher"
        )

    index_dir = get_index(bi_encoder_module, doc_datamodule, search_config)
    save_dir = tmp_path / "runs"
    search_callback = SearchCallback(search_config=search_config, index_dir=index_dir, save_dir=save_dir, use_gpu=False)

    trainer = LightningIRTrainer(
        logger=False,
        enable_checkpointing=False,
        callbacks=[search_callback],
        inference_mode=False,
    )
    trainer.test(bi_encoder_module, datamodule=query_datamodule)

    for dataloader in trainer.test_dataloaders:
        dataset = dataloader.dataset
        dataset_id = dataset.dataset_id.replace("/", "-")
        assert (save_dir / f"{dataset_id}.run").exists()
        run_df = pd.read_csv(
            save_dir / f"{dataset_id}.run",
            sep="\t",
            header=None,
            names=["query_id", "Q0", "doc_id", "rank", "score", "system"],
        )
        assert run_df["query_id"].nunique() == len(dataset)


def test_rerank_callback(tmp_path: Path, module: LightningIRModule, inference_datasets: Sequence[RunDataset]):
    datamodule = run_datamodule(module, inference_datasets)
    save_dir = tmp_path / "runs"
    rerank_callback = ReRankCallback(save_dir)
    trainer = LightningIRTrainer(logger=False, enable_checkpointing=False, callbacks=[rerank_callback])
    trainer.re_rank(module, datamodule)

    for dataloader in trainer.test_dataloaders:
        dataset = dataloader.dataset
        dataset_id = dataset.dataset_id.replace("/", "-")
        assert (save_dir / f"{dataset_id}.run").exists()
        run_df = pd.read_csv(
            save_dir / f"{dataset_id}.run",
            sep="\t",
            header=None,
            names=["query_id", "Q0", "doc_id", "rank", "score", "system"],
        )
        assert run_df["query_id"].nunique() == len(dataset)


def test_register_local_dataset_callback(model_name_or_path: str):
    callback = RegisterLocalDatasetCallback(
        dataset_id="test",
        docs=str(CORPUS_DIR / "docs.tsv"),
        queries=str(CORPUS_DIR / "queries.tsv"),
        qrels=str(CORPUS_DIR / "qrels.tsv"),
        docpairs=str(CORPUS_DIR / "docpairs.tsv"),
    )
    module = LightningIRModule(model_name_or_path=model_name_or_path, config=DprConfig(embedding_dim=4))
    datamodule = LightningIRDataModule(train_dataset=RunDataset("test"), train_batch_size=2)

    trainer = LightningIRTrainer(logger=False, enable_checkpointing=False, callbacks=[callback])

    trainer.test(module, datamodule)

    assert ir_datasets.registry._registered.get("test") is not None



"""Torch-based Sparse Searcher for Lightning IR Framework"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import torch

from ...modeling_utils.batching import _batch_pairwise_scoring
from ...models import SpladeConfig
from ..base.searcher import ExactSearchConfig, ExactSearcher
from .sparse_indexer import TorchSparseIndexConfig

if TYPE_CHECKING:
    from ...bi_encoder import BiEncoderEmbedding, BiEncoderModule


class TorchSparseIndex:
    """Torch-based sparse index for efficient retrieval."""

    def __init__(self, index_dir: Path, similarity_function: Literal["dot", "cosine"], use_gpu: bool = False) -> None:
        """Initialize the TorchSparseIndex.

        Args:
            index_dir (Path): Directory containing the index files.
            similarity_function (Literal["dot", "cosine"]): The similarity function to use.
            use_gpu (bool): Whether to use GPU for computations. Defaults to False.
        Raises:
            ValueError: If the similarity function is not recognized.
        """
        self.index = torch.load(index_dir / "index.pt", weights_only=True)
        self.config = TorchSparseIndexConfig.from_pretrained(index_dir)
        if similarity_function == "dot":
            self.similarity_function = self.dot_similarity
        elif similarity_function == "cosine":
            self.similarity_function = self.cosine_similarity
        else:
            raise ValueError("Unknown similarity function")
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    def score(self, embeddings: torch.Tensor) -> torch.Tensor:
        """Compute scores for the given embeddings.

        Args:
            embeddings (torch.Tensor): The embeddings to score.
        Returns:
            torch.Tensor: The computed scores.
        """
        embeddings = embeddings.to(self.index)
        similarity = self.similarity_function(embeddings, self.index).to_dense()
        return similarity

    @property
    def num_embeddings(self) -> int:
        """Get the number of embeddings in the index.

        Returns:
            int: The number of embeddings.
        """
        return self.index.shape[0]

    @staticmethod
    @_batch_pairwise_scoring
    @torch.autocast(device_type="cuda", enabled=False)
    def cosine_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute cosine similarity between two tensors.

        Args:
            x (torch.Tensor): The first tensor.
            y (torch.Tensor): The second tensor.
        Returns:
            torch.Tensor: The cosine similarity scores.
        """
        return y.matmul(x.T).T / (torch.norm(x, dim=-1)[:, None] * torch.norm(y, dim=-1)[None])

    @staticmethod
    @_batch_pairwise_scoring
    @torch.autocast(device_type="cuda", enabled=False)
    def dot_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute dot product similarity between two tensors.

        Args:
            x (torch.Tensor): The first tensor.
            y (torch.Tensor): The second tensor.
        Returns:
            torch.Tensor: The dot product similarity scores.
        """
        return y.matmul(x.T).T

    def to_gpu(self) -> None:
        """Move the index to GPU if available."""
        self.index = self.index.to(self.device)


class TorchSparseSearcher(ExactSearcher):
    """Torch-based sparse searcher for Lightning IR framework."""

    def __init__(
        self,
        index_dir: Path,
        search_config: TorchSparseSearchConfig,
        module: BiEncoderModule,
        use_gpu: bool = True,
    ) -> None:
        """Initialize the TorchSparseSearcher.

        Args:
            index_dir (Path): Directory containing the index files.
            search_config (TorchSparseSearchConfig): Configuration for the searcher.
            module (BiEncoderModule): The BiEncoder module to use for scoring.
            use_gpu (bool): Whether to use GPU for computations. Defaults to True.
        """
        self.search_config: TorchSparseSearchConfig
        self.index = TorchSparseIndex(index_dir, module.config.similarity_function, use_gpu)
        super().__init__(index_dir, search_config, module, use_gpu)
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    def to_gpu(self) -> None:
        """Move the searcher and index to GPU if available."""
        super().to_gpu()
        self.index.to_gpu()

    def _score(self, query_embeddings: BiEncoderEmbedding) -> torch.Tensor:
        if query_embeddings.scoring_mask is None:
            embeddings = query_embeddings.embeddings[:, 0]
        else:
            embeddings = query_embeddings.embeddings[query_embeddings.scoring_mask]
        scores = self.index.score(embeddings)
        return scores


class TorchSparseSearchConfig(ExactSearchConfig):
    """Configuration for the Torch-based sparse searcher."""

    search_class = TorchSparseSearcher
    SUPPORTED_MODELS = {SpladeConfig.model_type}



"""Configuration and model for SPLADE (SParse Lexical AnD Expansion) type models. Originally proposed in
`SPLADE: Sparse Lexical and Expansion Model for First Stage Ranking
<https://dl.acm.org/doi/abs/10.1145/3404835.3463098>`_.
"""

import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Literal, Self

import torch
from transformers import BatchEncoding

from ...bi_encoder import (
    BiEncoderEmbedding,
    BiEncoderTokenizer,
    SingleVectorBiEncoderConfig,
    SingleVectorBiEncoderModel,
)
from ...modeling_utils.embedding_post_processing import Pooler, Sparsifier
from ...modeling_utils.mlm_head import (
    MODEL_TYPE_TO_LM_HEAD,
    MODEL_TYPE_TO_OUTPUT_EMBEDDINGS,
    MODEL_TYPE_TO_STATE_DICT_KEY_MAPPING,
    MODEL_TYPE_TO_TIED_WEIGHTS_KEYS,
)


class SpladeConfig(SingleVectorBiEncoderConfig):
    """Configuration class for a SPLADE model."""

    model_type = "splade"
    """Model type for a SPLADE model."""

    def __init__(
        self,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        similarity_function: Literal["cosine", "dot"] = "dot",
        sparsification_strategy: Literal["relu", "relu_log", "relu_2xlog"] | None = "relu_log",
        pooling_strategy: Literal["first", "mean", "max", "sum"] = "max",
        query_weighting: Literal["contextualized", "static"] | None = "contextualized",
        query_expansion: bool = True,
        doc_weighting: Literal["contextualized", "static"] | None = "contextualized",
        doc_expansion: bool = True,
        **kwargs,
    ) -> None:
        """A SPLADE model encodes queries and documents separately. Before computing the similarity score, the
        contextualized token embeddings are projected into a logit distribution over the vocabulary using a pre-trained
        masked language model (MLM) head. The logit distribution is then sparsified and aggregated to obtain a single
        embedding for the query and document.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            similarity_function (Literal["cosine", "dot"]): Similarity function to compute scores between query and
                document embeddings. Defaults to "dot".
            sparsification_strategy (Literal['relu', 'relu_log', 'relu_2xlog'] | None): Whether and which sparsification
                function to apply. Defaults to None.
            pooling_strategy (Literal["first", "mean", "max", "sum"]): Pooling strategy for query embeddings.
                Defaults to "max".
            query_weighting (Literal["contextualized", "static"] | None): Whether to reweight query embeddings.
                Defaults to "contextualized".
            query_expansion (bool): Whether to allow query expansion. Defaults to True.
            doc_weighting (Literal["contextualized", "static"] | None): Whether to reweight document embeddings.
                Defaults to "contextualized".
            doc_expansion (bool): Whether to allow document expansion. Defaults to True.
        """
        super().__init__(
            query_length=query_length,
            doc_length=doc_length,
            similarity_function=similarity_function,
            sparsification_strategy=sparsification_strategy,
            pooling_strategy=pooling_strategy,
            **kwargs,
        )
        if query_expansion and not query_weighting:
            raise ValueError("If query_expansion is True, query_weighting must also be True.")
        if doc_expansion and not doc_weighting:
            raise ValueError("If doc_expansion is True, doc_weighting must also be True.")
        self.query_weighting = query_weighting
        self.query_expansion = query_expansion
        self.doc_weighting = doc_weighting
        self.doc_expansion = doc_expansion

    @property
    def embedding_dim(self) -> int:
        vocab_size = getattr(self, "vocab_size", None)
        if vocab_size is None:
            raise ValueError("Unable to determine embedding dimension.")
        return vocab_size

    @embedding_dim.setter
    def embedding_dim(self, value: int) -> None:
        pass


class SpladeModel(SingleVectorBiEncoderModel):
    """Sparse lexical SPLADE model. See :class:`SpladeConfig` for configuration options."""

    config_class = SpladeConfig
    """Configuration class for a SPLADE model."""

    def __init__(self, config: SingleVectorBiEncoderConfig, *args, **kwargs) -> None:
        """Initializes a SPLADE model given a :class:`SpladeConfig`.

        Args:
            config (SingleVectorBiEncoderConfig): Configuration for the SPLADE model.
        """
        super().__init__(config, *args, **kwargs)
        # grab language modeling head based on backbone model type
        layer_cls = MODEL_TYPE_TO_LM_HEAD[config.backbone_model_type or config.model_type]
        self.projection = layer_cls(config)
        tied_weight_keys = getattr(self, "_tied_weights_keys", []) or []
        tied_weight_keys = tied_weight_keys + [
            f"projection.{key}"
            for key in MODEL_TYPE_TO_TIED_WEIGHTS_KEYS[config.backbone_model_type or config.model_type]
        ]
        self._tied_weights_keys = tied_weight_keys
        self.query_weights = None
        if config.query_weighting == "static":
            self.query_weights = torch.nn.Embedding(config.vocab_size, 1)
        self.doc_weights = None
        if config.doc_weighting == "static":
            self.doc_weights = torch.nn.Embedding(config.vocab_size, 1)

        self.pooler = Pooler(config.pooling_strategy)
        self.sparsifier = Sparsifier(config.sparsification_strategy)

    def encode(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> BiEncoderEmbedding:
        """Encodes a batched tokenized text sequences and returns the embeddings and scoring mask.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BiEncoderEmbedding: Embeddings and scoring mask.
        """
        weighting = getattr(self.config, f"{input_type}_weighting")
        expansion = getattr(self.config, f"{input_type}_expansion")
        token_mask = None
        if weighting is None or weighting == "static" or not expansion:
            token_mask = torch.zeros(
                encoding["input_ids"].shape[0],
                self.config.embedding_dim,
                device=encoding["input_ids"].device,
                dtype=torch.float32,
            )
            if weighting == "static":
                weights = getattr(self, f"{input_type}_weights")(encoding["input_ids"]).squeeze(-1)
            else:
                weights = encoding["attention_mask"].to(token_mask)
            weights = weights.masked_fill(~(encoding["attention_mask"].bool()), 0.0)
            token_mask = token_mask.scatter(1, encoding["input_ids"], weights)
            if weighting is None or weighting == "static":
                # inference free
                return BiEncoderEmbedding(token_mask[:, None], None, encoding)

        embeddings = self._backbone_forward(**encoding).last_hidden_state
        embeddings = self.projection(embeddings)
        embeddings = self.sparsifier(embeddings)
        embeddings = self.pooler(embeddings, encoding["attention_mask"])
        if token_mask is not None:
            embeddings = embeddings * token_mask[:, None]
        return BiEncoderEmbedding(embeddings, None, encoding)

    @classmethod
    def from_pretrained(cls, model_name_or_path: str | Path, *args, **kwargs) -> Self:
        """Loads a pretrained model and handles mapping the MLM head weights to the projection head weights. Wraps
        the transformers.PreTrainedModel.from_pretrained_ method to return a derived LightningIRModel.
        See :class:`LightningIRModelClassFactory` for more details.

.. _transformers.PreTrainedModel.from_pretrained: \
    https://huggingface.co/transformers/main_classes/model.html#transformers.PreTrainedModel.from_pretrained

        .. ::doctest
        .. highlight:: python
        .. code-block:: python

            >>> # Loading using model class and backbone checkpoint
            >>> type(CrossEncoderModel.from_pretrained("bert-base-uncased"))
            <class 'lightning_ir.base.class_factory.CrossEncoderBertModel'>
            >>> # Loading using base class and backbone checkpoint
            >>> type(LightningIRModel.from_pretrained("bert-base-uncased", config=CrossEncoderConfig()))
            <class 'lightning_ir.base.class_factory.CrossEncoderBertModel'>

        Args:
            model_name_or_path (str | Path): Name or path of the pretrained model.
        Returns:
            Self: A derived LightningIRModel consisting of a backbone model and a LightningIRModel mixin.
        Raises:
            ValueError: If called on the abstract class :class:`SpladeModel` and no config is passed.
        """
        key_mapping = kwargs.pop("key_mapping", {})
        config = cls.config_class
        # map mlm projection keys
        model_type = config.backbone_model_type or config.model_type
        if model_type in MODEL_TYPE_TO_STATE_DICT_KEY_MAPPING:
            key_mapping.update(MODEL_TYPE_TO_STATE_DICT_KEY_MAPPING[model_type])
        if not key_mapping:
            warnings.warn(
                f"No mlm key mappings for model_type {model_type} were provided. "
                "The pre-trained mlm weights will not be loaded correctly.",
                stacklevel=2,
            )
        model = super().from_pretrained(model_name_or_path, *args, key_mapping=key_mapping, **kwargs)
        return model

    def set_output_embeddings(self, new_embeddings: torch.nn.Module) -> None:
        if self.config.projection == "mlm":
            raise NotImplementedError("Setting output embeddings is not supported for models with MLM projection.")
            # TODO fix this (not super important, only necessary when additional tokens are added to the model)
            # module_names = MODEL_TYPE_TO_OUTPUT_EMBEDDINGS[self.config.backbone_model_type or self.config.model_type]
            # module = self
            # for module_name in module_names.split(".")[:-1]:
            #     module = getattr(module, module_name)
            # setattr(module, module_names.split(".")[-1], new_embeddings)
            # setattr(module, "bias", new_embeddings.bias)

    def get_output_embeddings(self) -> torch.nn.Module | None:
        """Returns the output embeddings of the model for tieing the input and output embeddings. Returns None if no
        MLM head is used for projection.

        Returns:
            torch.nn.Module | None: Output embeddings of the model.
        """
        module_names = MODEL_TYPE_TO_OUTPUT_EMBEDDINGS[self.config.backbone_model_type or self.config.model_type]
        output = self.projection
        for module_name in module_names.split("."):
            output = getattr(output, module_name)
        return output


class SpladeTokenizer(BiEncoderTokenizer):
    """Tokenizer class for SPLADE models."""

    config_class = SpladeConfig
    """Configuration class for a SPLADE model."""

    def __init__(
        self,
        *args,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        add_marker_tokens: bool = False,
        query_weighting: Literal["contextualized", "static"] = "contextualized",
        doc_weighting: Literal["contextualized", "static"] = "contextualized",
        **kwargs,
    ):
        super().__init__(
            *args, query_length=query_length, doc_length=doc_length, add_marker_tokens=add_marker_tokens, **kwargs
        )
        """Initializes a SPLADE model's tokenizer. Encodes queries and documents separately. Optionally adds
        marker tokens to encoded input sequences.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            add_marker_tokens (bool): Whether to add extra marker tokens [Q] / [D] to queries / documents.
                Defaults to False.
            query_expansion (bool): Whether to expand queries with mask tokens. Defaults to False.
            attend_to_query_expanded_tokens (bool): Whether to allow query tokens to attend to mask expanded query
                tokens. Defaults to False.
            doc_expansion (bool): Whether to expand documents with mask tokens. Defaults to False.
            attend_to_doc_expanded_tokens (bool): Whether to allow document tokens to attend to mask expanded document
                tokens. Defaults to False.
            query_weighting (Literal["contextualized", "static"]): Whether to apply weighting to query tokens.
                Defaults to "contextualized".
            doc_weighting (Literal["contextualized", "static"]): Whether to apply weighting to document tokens.
                Defaults to "contextualized".
        Raises:
            ValueError: If `add_marker_tokens` is True and a non-supported tokenizer is used.
        """
        super().__init__(
            *args,
            query_length=query_length,
            doc_length=doc_length,
            add_marker_tokens=add_marker_tokens,
            query_weighting=query_weighting,
            doc_weighting=doc_weighting,
            **kwargs,
        )
        self.query_weighting = query_weighting
        self.doc_weighting = doc_weighting

    def tokenize_input_sequence(
        self, text: Sequence[str] | str, input_type: Literal["query", "doc"], *args, **kwargs
    ) -> BatchEncoding:
        """Tokenizes an input sequence. This method is used to tokenize both queries and documents.

        Args:
            text (Sequence[str] | str): Input text to tokenize.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BatchEncoding: Tokenized input sequences.
        """
        post_processer = getattr(self, f"{input_type}_post_processor")
        kwargs["max_length"] = getattr(self, f"{input_type}_length")
        weighting = getattr(self, f"{input_type}_weighting")
        if weighting is None or weighting == "static":
            kwargs["add_special_tokens"] = False
        if "padding" not in kwargs:
            kwargs["truncation"] = True
        return self._encode(text, *args, post_processor=post_processer, **kwargs)
